#
# utilfns.tcl
#
# A bunch of little utility subroutines for tcl
#
# dpwe@media.mit.edu 1995aug06
#

proc quit {} {
    exit 
}

proc WinWidth {win} {
    # figure the width of a window's interior (work area) by asking 
    # the window manager, then knocking off the borderwidth
    set bw [$win cget -borderwidth]
    set w  [winfo width $win]
    # knock off two lots of $bw, one for each side
    return [expr "$w - 2*$bw"]
}

proc WinHeight {win} {
    # figure the height of a window's interior (work area) by asking 
    # the window manager, then knocking off the borderwidth
    set bw [$win cget -borderwidth]
    set h  [winfo height $win]
    # knock off two lots of $bw, one for each side
    return [expr "$h - 2*$bw"]
}

proc CanvTextDims {cnv {text "N"} {font {}}} {
    # return the w and h for the given text & font if drawn in the given 
    # canvas.  If omitted, font uses default canvas font; 
    # With no text specified, returns the size of an upper-case N.
    if { $font == "" }  {
	# no font specified, allow default to happen
	set tag [$cnv create text 10 10 -text $text]
    } else {
	set tag [$cnv create text 10 10 -text $text -font $font]
    }
    set bbx [$cnv bbox $tag]
    $cnv delete $tag
    return [list [expr "[lindex $bbx 2]-[lindex $bbx 0]"] \
	    [expr "[lindex $bbx 3]-[lindex $bbx 1]"]]
}

proc max args {
    # Treats args as a list of numbers, returns the largest
    set list [eval "concat $args"]
    set rslt [lindex $list 0]
    foreach el $list {
	if { $el > $rslt } {
	    set rslt $el
	}
    }
    return $rslt
}

proc min args {
    # Treats args as a list of numbers, returns the smallest
    set list [eval "concat $args"]
    set rslt [lindex $list 0]
    foreach el $list {
	if { $el < $rslt } {
	    set rslt $el
	}
    }
    return $rslt
}

proc sum args {
    # Treats args as a list of reals, returns the sum
    set list [eval "concat $args"]
    # A decimal point in the list means treat them as reals
    if {[string first "." $list] >= 0} {
	set rslt 0.0
    } else {
	set rslt 0
    }
    foreach el $list {
	set rslt [expr $rslt+$el]
    }
    return $rslt
}

proc lwithout {args} {
    # usage: lwithout ?-exact|-glob? list val
    # Remove all instances of $val from $list; return the remainder
    lassign [processProcArgs $args] match list val
    if {$match == ""} {set match -exact}
    while { [set ix [lsearch $match $list $val]] >= 0 } {
        # null out the $ix'th value
        set list [lreplace $list $ix $ix]
    }
    return $list
}

proc lwithoutl {args} {
    # usage: lwithout ?-exact|-glob? list stoplist
    # Remove all instances of els of $stoplist from $list; return the remainder
#    set outlist {}
#    foreach el $list {
#	if {[lsearch $match $stoplist $el]==-1} {
#	    lappend outlist $el
#	}
#    }
    lassign [processProcArgs $args] match list stoplist
    if {$match == ""} {set match -exact}
    set outlist $list
    foreach stop $stoplist {
	set outlist [lwithout $match $outlist $stop]
    }
    return $outlist
}

proc lmap {list fn} {
    # Return a new list where each element is the result of applying 
    # $fn to an element of $list
    set out {}
    foreach el $list {
	lappend out [eval "$fn $el"]
    }
    return $out
}

proc lmapl {list fn} {
    # Return a new list where each element is the result of applying 
    # $fn to an element of $list - but those elements form a single arg
    set out {}
    foreach el $list {
	lappend out [eval "$fn {$el}"]
    }
    return $out
}

proc nth {n args} {
    # Quick function to extract the nth element from a list.  
    # List is the last arg, thus can be used as a prefix fn 
    # for lmap
    # lists likely to have been split into args by eval
    set list [eval "list $args"]
    return [lindex $list $n]
}

proc cons {head rest} {
    # Return a list whose car is $head and whose cdr is $rest
    set out [concat [list $head] $rest]
}

proc reverse list {
    # Return the list with its top-level elements in reverse order
    set out {}
    foreach e $list {
	set out [cons $e $out]
    }
    return $out
}

proc resetArray {name} {
    # from original itcl examples
    uplevel 1 "catch \"unset $name\""
    uplevel 1 "set ${name}(0) \"make-this-an-array\""
    uplevel 1 "unset ${name}(0)"
}

proc getMeanSd {data {ignorevals {{}}}} {
    # $data is a list of numbers; calculate their mean and sd
    # return [list $mean $sd $n]
    # Values from $ignorvals will not be included
    set sum 0.0
    set sumsq 0.0
    set n 0
    foreach x $data {
	# maybe reject certain elements
	if {[lsearch -exact $ignorevals $x]==-1} {
	    set sum   [expr $sum+$x]
	    set sumsq [expr $sumsq + ($x*$x)]
	    incr n
	}
    }
    if {$n > 0} {
	set mean [expr $sum/$n]
	# rounding problems with tcl
	set sqmean [expr $mean*$mean]
	set meansq [expr $sumsq/$n]
	if {$sqmean >= $meansq} {
	    set sd 0.0
	} else {
	    set sd [expr sqrt($meansq - $sqmean)]
	}
    } else {
	set sd 0.0
	set mean 0.0
    }
    return [list $mean $sd $n]
}

proc lassign {list args} {
    # take a list and some number of variable names; set the 
    # variables to the corresponding elements of the list
    set i 0
    foreach arg $args {
	set v [lindex $list $i]
	#regsub -all {([^\])([{}])} $v {\1\\\2} v
	#uplevel 1 "set $arg \{$v\}"
	uplevel 1 "set $arg [list $v]"
	incr i
    }
}


proc car {list} {
    # lisp's car - first element of a list
    return [lindex $list 0]
}

proc cdr {list} {
    # lisp's cdr - list without first element
    return [lrange $list 1 e]
}

proc last {list} {
    # return the last top-level element of a list
    return [lindex $list [expr [llength $list]-1]]
}

proc uniqueCommand {stem {startFrom 0}} {
    # append increasing integers to $stem until we find one that isn't
    # already a command.  Integers start search from $startFrom
    set existingCmds [info commands ${stem}*]
    set i $startFrom
    set found 0
    while {!$found} {
	set cmdName ${stem}$i
	if {[lsearch -exact $existingCmds $cmdName]==-1} {
	    set found 1
	}
	incr i
    }
    return $cmdName
}

proc getenv {envvar {dflt ""}} {
    # Return the environment variable, or $dflt if it doesn't exist
    global env
    if {[info exists env($envvar)]} {
	return $env($envvar)
    } else {
	return $dflt
    }
}

proc processProcArgs {arglist} {
    # Return the list $args parsed into a sublist of "-"-prefixed switches 
    # and remaining regular args
    set switches ""
    set otherargs ""
    set moreswitches 1
    set nargs [llength $arglist]
    for {set i 0} {$i<$nargs} {incr i} {
	if {$moreswitches && [string index [lindex $arglist $i] 0] == "-"} {
	    lappend switches [lindex $arglist $i]
	} else {
	    lappend otherargs [lindex $arglist $i]
	    # prevent subsequent arg elements from being taken as switches
	    set moreswitches 0
	}
    }
    return [cons $switches $otherargs]
}

proc regmod {args} {
    # args are [-switches] pattern string replacement
    # Perform regsub and return the result
    # Separate switches from other args
    lassign [processProcArgs $args] switches pattern string repl
    # Result will be just the string if regsub fails
    set rslt $string
    eval regsub $switches \{$pattern\} \{$string\} \{$repl\} rslt
    return $rslt
}

proc stringreplace {string first last replacement} {
    # Like list replace, replace a certain range of chrs in a string
    return "[string range $string 0 [expr $first-1]]$replacement[string range $string [expr $last+1] e]"
}

proc GetCLVars {argv} {
    # Take a series of command line args of form var=val
    # and set the corresponding Tcl global variables
    # Return the names of the set variables.
    set varnames {}
    foreach m $argv {
	set varconts [split $m "="]
	if {[llength $varconts]==2} {
	    set varname [lindex $varconts 0]
	    global $varname
	    set $varname [lindex $varconts 1]
	    lappend varnames $varname
	} else {
	    puts stderr "GetCLVars: couldn't parse '$m' as 'var=val'"
	}
    }
    return $varnames
}

proc _compareCarNum {a b} {
    # For sorting {count token} lists, treat inputs as lists whose 
    # first element is numeric, and compare just those first 
    # elements numerically
    set n1 [lindex $a 0]
    set n2 [lindex $b 0]
    if {$n1 < $n2} {
	return -1
    } elseif {$n1 > $n2} {
	return 1
    } else {
	return 0
    }
}

proc UniqFreq {list} {
    # Take a list of tokens and return a list of the unique ones 
    # alongside their frequencies of occurrence.  Thus 
    # "a b a c c a d" gives "{3 a} {2 c} {1 b} {1 d}"
    set slist [lsort $list]
    set op ""
    set lastel [lindex $slist 0]
    set count 1
    foreach el [lrange $slist 1 e] {
	if {$el != $lastel} {
	    lappend op [list $count $lastel]
	    set count 0
	}
	set lastel $el
	incr count
    }
    lappend op [list $count $lastel]
    return [lsort -command _compareCarNum -decreasing $op]
}

