//
// tclAudIO.cc
// 
// tcl extension to interface to sound input/output
// 
// dpwe@media.mit.edu 1995jan31
// uses my audIO.h library
//

#include <assert.h>
#include <math.h>	// for sqrt in scan (rms)
extern "C" {
#include "tcl.h"
#include "audIO.h"

Tcl_HashTable audioTable;		// Convert audio names to ptrs

#define DFLT_CHANS	1
#define DFLT_SRATE	22050
#define DFLT_FORMAT	AFMT_SHORT
#define DFLT_MODE	'r'

typedef struct {
    double samplerate;
    char   mode;
    int    format;
    int    chans;
    AUSTRUCT *au;
} Audio;

int Audio_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int audioCommand(ClientData clientData, Tcl_Interp *interp, 
		 int argc, char **argv);
int audioProc(ClientData clientData, Tcl_Interp *interp, 
	      int argc, char **argv);
int audioConfigure(Audio *audioPtr, Tcl_Interp *interp, 
	      char *cmdName, int argc, char **argv);
void audioDelete(ClientData clientData);
}

/* Usage:
 * audio auObj -mode {read|write} -samplerate 44100 -chans 1 -format short
 * auObj config -mode r -samplerate 44100 -chans 1 -format short
 * 
 * auObj read <frames>
 * auObj write <frames>
 * auObj scan <nframes>     reads <nframes> and returns their energy
 */

int Audio_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tcl_CreateCommand(interp, "audio", audioCommand, (ClientData) NULL,
	    (void (*)(ClientData)) NULL);
    // initialize the hash table
    Tcl_InitHashTable(&audioTable, TCL_STRING_KEYS);
    return TCL_OK;
}

int audioCommand(ClientData clientData, Tcl_Interp *interp, 
		 int argc, char **argv)
{	/* tcl command to instance a new audio object */
    if(argc < 2) {
	Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
    // allocate a new object
    Audio *audioPtr = new Audio;
    audioPtr->au = NULL;
    // default values
    audioPtr->chans      = DFLT_CHANS;
    audioPtr->samplerate = DFLT_SRATE;
    audioPtr->format     = DFLT_FORMAT;
    audioPtr->mode       = DFLT_MODE;
    // put it into the tcl command namespace
    Tcl_CreateCommand(interp, argv[1], audioProc, (ClientData)audioPtr, 
		      audioDelete);
    // .. and add it as a key in the hash table lookup
    int isnew;
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_CreateHashEntry(&audioTable, argv[1], &isnew);
    if(!isnew)  {		// re-using a name - must destroy old command
	// actually, command (and hence data structure) will have been 
	// deleted when we re-used the command name above, so it 
	// doesn't matter
    }
    Tcl_SetHashValue(entryPtr, audioPtr);

    // pass rest of command to configure
    if(audioConfigure(audioPtr, interp, argv[0], argc-2, argv+2) != TCL_OK)  {
	return TCL_ERROR;
    }
    interp->result = argv[1];
    return TCL_OK;
}

int audioProc(ClientData clientData, Tcl_Interp *interp, 
	      int argc, char **argv)
{	/* handle the commands supported by the audioObj */
    Audio *au = (Audio *)clientData;
    int		result = TCL_OK;
    char	c;
    int		length;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if((c=='c') && strncmp(argv[1], "config", length)==0)  {
	return audioConfigure(au, interp, argv[0], argc-1, argv+1);
    } else if( (c=='r') && strncmp(argv[1], "read", length)==0)  {
	int frames;
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " read <framecount>\"", NULL);
	    return TCL_ERROR;
	}
	if(Tcl_GetInt(interp, argv[2], &frames)!=TCL_OK)
	    return TCL_ERROR;
	int chans = au->chans;
	if(au->format==AFMT_SHORT)  {
	    short *s = new short[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = AURead(au->au, s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%d ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans-1; ++ch) {
			sprintf(str, "%d ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%d", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else if(au->format==AFMT_FLOAT)  {
	    float *s = new float[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = AURead(au->au, s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%.6f ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans-1; ++ch) {
			sprintf(str, "%.6f ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%.6f", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else if(au->format==AFMT_LONG)  {
	    long *s = new long[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = AURead(au->au, s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%ld ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans; ++ch) {
			sprintf(str, "%ld ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%ld", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else {
	    Tcl_AppendResult(interp, 
			     "hopelessly confused about format in read");
	    return TCL_ERROR;
	}
//	sprintf(interp->result, "(read %d frames)", frames);
    } else if( (c=='w') && strncmp(argv[1], "write", length)==0)  {
	char **listels;
	int nlistels;

	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " write <datalist>\"", NULL);
	    return TCL_ERROR;
	}
	Tcl_SplitList(interp, argv[2], &nlistels, &listels);
	// process in here
	free(listels);
	sprintf(interp->result, "(wrote %d frames)", nlistels);
    } else if( (c=='s') && strncmp(argv[1], "scan", length)==0)  {
	int frames;
	int chans = au->chans;

	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " scan nframes\"", NULL);
	    return TCL_ERROR;
	}
	if(Tcl_GetInt(interp, argv[2], &frames)!=TCL_OK)
	    return TCL_ERROR;
	assert(au->format == AFMT_SHORT);
	short *buf = new short[chans * frames];
	short *sptr = buf;
	int got = AURead(au->au, buf, frames);
	float *sums = new float[chans];
	float d;
	int i,j;
	for(i=0; i<chans;++i) 	sums[i] = 0;
	for(i=0; i<frames; ++i)  {
	    for(j=0;j<chans;++j)  {
		d = (*sptr++)/32768.0;
		sums[j] += d*d;
	    }
	}
	char s[32];
	for(i=0; i<chans;++i) {
	    sprintf(s, "%.6f", sqrt(sums[i]/frames));
	    Tcl_AppendResult(interp, s, NULL);
	}
	delete[] sums;
	delete[] buf;
    } else {
	Tcl_AppendResult(interp, "bad command \"", argv[1], 
		"\":  config/read/write/scan",
			 (char *) NULL);
	result = TCL_ERROR;
    }
    return result;
}

int audioConfigure(Audio *au, Tcl_Interp *interp, 
		   char *cmdName, int argc, char **argv)
{	/* Handle arguments specifically to configure */
    int i = -1; 
    char c, *s;
    int len;

    while(++i < argc)  {
	// fprintf(stderr, "argv[%d]='%s'\n", i, argv[i]);
	if(argv[i][0] != '-')  {
	    Tcl_AppendResult(interp, "configure error: should be \"", 
			     cmdName, " config {-mode|-samplerate|-chans|-format|-nframes|-filename} ?val?", NULL);
	    return TCL_ERROR;
	}
	s = &(argv[i][1]);  c = *s;  len = strlen(s);
	if( (c=='m') && strncmp(s,"mode",len)==0)  {
	    char mc;
	    ++i;
	    if(i==argc)  {
		mc = au->mode;
	    } else {
		char *arg = argv[i];
		int larg = strlen(arg);
		char argi = *arg;
		if( (argi=='r') && strncmp(arg, "read", larg)==0)  {
		    mc = au->mode = 'r';
		} else if( (argi=='w') && strncmp(arg, "write", larg)==0)  {
		    mc = au->mode = 'w';
		} else {
		    Tcl_AppendResult(interp, "error: should be \"", 
				     cmdName, " mode {read|write}\"", NULL);
		    return TCL_ERROR;
		}
	    }
	    if(mc=='r')
		sprintf(interp->result, "read");
	    else
		sprintf(interp->result, "write");
	} else if( (c=='s' && strncmp(s,"samplerate",len)==0) ) {
	    double srate;
	    ++i;
	    if(i==argc)  {
		srate = au->samplerate;
	    } else {
		if(Tcl_GetDouble(interp, argv[i], &srate)!=TCL_OK)
		    return TCL_ERROR;
		 au->samplerate = srate;
	    }
	    sprintf(interp->result, "%.2f", srate);
	} else if( (c=='c' && strncmp(s,"chans",len)==0) ) {
	    int chans;
	    ++i;
	    if(i==argc)  {
		chans = au->chans;
	    } else {
		if(Tcl_GetInt(interp, argv[i], &chans)!=TCL_OK)
		    return TCL_ERROR;
		au->chans = chans;
	    }
	    sprintf(interp->result, "%d", chans);
	} else if( (c=='f') && strncmp(s,"format",len)==0)  {
	    int fmt;
	    ++i;
 	    if(i==argc)  {
		fmt = au->format;
	    } else {
		char *arg = argv[i];
		int larg = strlen(arg);
		char argi = *arg;
		if( (argi=='s') && strncmp(arg, "short", larg)==0)  {
		    fmt = AFMT_SHORT;
		} else if( (argi=='l') && strncmp(arg, "long", larg)==0)  {
		    fmt = AFMT_LONG;
		} else if( (argi=='f') && strncmp(arg, "float", larg)==0)  {
		    fmt = AFMT_FLOAT;
		} else {
		    Tcl_AppendResult(interp, "error: should be \"", 
				     cmdName, " format {short|long|float}\"", NULL);
		    return TCL_ERROR;
		}
		au->format = fmt;
	    }
	    sprintf(interp->result, (fmt==AFMT_SHORT)?"short":((fmt==AFMT_LONG)?"long":((fmt==AFMT_FLOAT)?"float":"unknown")));
	} else {
	    Tcl_AppendResult(interp, "unrecognized: should be \"", 
			     cmdName, "config {-mode|-samplerate|-chans|-format} ?val?", NULL);
	    return TCL_ERROR;
	}
    }
    // now actually open the audio if it's not already
    if(au->au==NULL)  {
	au->au = AUOpen(au->format, au->samplerate, au->chans, 0,
			(au->mode=='r')?AIO_INPUT:AIO_OUTPUT);
	if(au->au==NULL)  {
	    Tcl_AppendResult(interp, "error opening audio output", NULL);
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

void audioDelete(ClientData clientData)
{	/* a particular audioObj is being deleted */
    Audio *au = (Audio *)clientData;

    AUClose(au->au,0);
    delete au;
}
    
