//
// tclFloatArray.cc
// 
// tcl extension to have 2D arrays of floats as tcl 'object-oriented' commands
// 
// dpwe@media.mit.edu 1995may09 after tclSound.cc
// uses my floatArray.cc object.
//

#include <assert.h>
#include <string.h>

#include "floatArray.h"
extern "C" {
#include "tcl.h"

Tcl_HashTable farrayTable;		// Convert sound names to ptrs

int Farray_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int farrayCommand(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv);
int farrayProc(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
int farrayConfigure(floatArray *faPtr, Tcl_Interp *interp, 
	      char *cmdName, int argc, char **argv);
void farrayDelete(ClientData clientData);
}

int Farray_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tcl_CreateCommand(interp, "farray", farrayCommand, (ClientData) NULL,
	    (void (*)(ClientData)) NULL);
    // initialize the hash table
    Tcl_InitHashTable(&farrayTable, TCL_STRING_KEYS);
    return TCL_OK;
}

int farrayCommand(ClientData /* clientData */, Tcl_Interp *interp, 
		  int argc, char **argv)
{	/* tcl command to instance a new sound object */
    if(argc < 2) {
	Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
    // allocate a new object
    floatArray *faPtr = new floatArray(0,0);
    // put it into the tcl command namespace
    Tcl_CreateCommand(interp, argv[1], farrayProc, (ClientData)faPtr, 
		      farrayDelete);
    // .. and add it as a key in the hash table lookup
    int isnew;
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_CreateHashEntry(&farrayTable, argv[1], &isnew);
    if(!isnew)  {		// re-using a name - must destroy old command
	// actually, command (and hence data structure) will have been 
	// deleted when we re-used the command name above, so it 
	// doesn't matter
    }
    Tcl_SetHashValue(entryPtr, faPtr);

    // pass rest of command to configure
    if(farrayConfigure(faPtr, interp, argv[0], argc-2, argv+2) != TCL_OK)  {
	return TCL_ERROR;
    }
    interp->result = argv[1];
    return TCL_OK;
}

int farrayProc(ClientData clientData, Tcl_Interp *interp, 
	      int argc, char **argv)
{	/* handle the commands supported by the floatArray obj */
    floatArray *faPtr = (floatArray *)clientData;
    int		result = TCL_OK;
    char	c;
    int		length;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if((c=='c') && strncmp(argv[1], "configure", length)==0)  {
	return farrayConfigure(faPtr, interp, argv[0], argc-1, argv+1);
    }
    if( (c=='l') && strncmp(argv[1], "load", length)==0)  {
	int dataset = 0;
	if(argc < 3 || argc > 4)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " load <filename> ?dataset?\"", NULL);
	    return TCL_ERROR;
	}
	if(argc>3)  {
	    if(Tcl_GetInt(interp, argv[3], &dataset)!=TCL_OK)
		return TCL_ERROR;
	}
	if(faPtr->loadDAT(argv[2], dataset)==0)  {
	    Tcl_AppendResult(interp, "cannot read from DAT file '", 
			     argv[2], "'", NULL);
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    if( (c=='s') && strncmp(argv[1], "save", length)==0)  {
	int dataset = 0;
	int trunc = 1;
	if(argc < 3 || argc > 5)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " save <filename> ?dataset? ?trunc?\"", 
			     NULL);
	    return TCL_ERROR;
	}
	if(argc>3)  {
	    if(Tcl_GetInt(interp, argv[3], &dataset)!=TCL_OK)
		return TCL_ERROR;
	}
	if(argc>4)  {
	    if(Tcl_GetInt(interp, argv[4], &trunc)!=TCL_OK)
		return TCL_ERROR;
	}
	if(faPtr->saveDAT(argv[2], dataset, trunc)==0)  {
	    Tcl_AppendResult(interp, "cannot write to DAT file '", 
			     argv[2], "'", NULL);
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    if( (c=='c') && strncmp(argv[1], "clear", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " clear\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->Clear();
    } else if( (c=='t') && strncmp(argv[1], "transpose", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " transpose\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->Transpose();
    } else if( (c=='s') && strncmp(argv[1], "size", length)==0)  {
	if(argc!=2 && argc!=4)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " size ?rows cols?\"", NULL);
	    return TCL_ERROR;
	}
	if(argc == 2)  {	/* without new size, just report size */
	    sprintf(interp->result, "%d %d", faPtr->Rows(), faPtr->Cols());
	    return TCL_OK;
	} else {	/* argc==4 i.e. new size specified */
	    int rows, cols;
	    if(Tcl_GetInt(interp, argv[2], &rows)!=TCL_OK)
		return TCL_ERROR;
	    if(Tcl_GetInt(interp, argv[3], &cols)!=TCL_OK)
		return TCL_ERROR;
	    faPtr->Resize(rows, cols);
	    return TCL_OK;
	}
    } else if( (c=='v') && strncmp(argv[1], "val", length)==0)  {
	if(argc<4 || (argc&1)==1)  {	/* must be even number & at least 2 */
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " val r1 c1 ?r2 c2 ...?\"", NULL);
	    return TCL_ERROR;
	}
	int npairs = (argc-2)/2;
	int i;
	int r,c;
	char s[32];
	for(i = 0; i<npairs; ++i)  {
	    if(Tcl_GetInt(interp, argv[2+2*i], &r)!=TCL_OK)  {
		Tcl_ResetResult(interp);
		return TCL_ERROR;
	    }
	    if(Tcl_GetInt(interp, argv[3+2*i], &c)!=TCL_OK)  {
		Tcl_ResetResult(interp);
		return TCL_ERROR;
	    }
	    if(r >= faPtr->Rows() || c >= faPtr->Cols())  {
		char s1[32], s2[32];
		sprintf(s1, "(%d,%d)", r, c);
		sprintf(s2, "(%d,%d)", faPtr->Rows(), faPtr->Cols());
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "point ", s1, " is not within ", 
				 s2, NULL);
		return TCL_ERROR;
	    }
	    sprintf(s, "%.3f ", faPtr->Val(r, c));
	    Tcl_AppendResult(interp, s, NULL);
	}
	return TCL_OK;
    } else if( (c=='c') && strncmp(argv[1], "copy", length)==0)  {
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " copy <farray>\"", NULL);
	    return TCL_ERROR;
	}	
	floatArray *srcFaPtr;
	Tcl_HashEntry *entryPtr;
	entryPtr = Tcl_FindHashEntry(&farrayTable, argv[2]);
	if(entryPtr == NULL) {  // no farray registered to that name
	    Tcl_AppendResult(interp, "no such farray: ", argv[2], NULL);
	    return TCL_ERROR;
	}
	srcFaPtr = (floatArray *)Tcl_GetHashValue(entryPtr);
	faPtr->Copy(*srcFaPtr);
	return TCL_OK;
    } else if( (c=='p') && strncmp(argv[1], "print", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " print\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->Print();
	return TCL_OK;
    } else if( (c=='n') && strncmp(argv[1], "normrows", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " normrows\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->NormRows();
	return TCL_OK;
    } else if( (c=='n') && strncmp(argv[1], "normcols", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " normcols\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->NormCols();
	return TCL_OK;
    } else if( (c=='n') && strncmp(argv[1], "norm", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " norm\"", NULL);
	    return TCL_ERROR;
	}
	faPtr->Norm();
	return TCL_OK;
    } else if( (c=='m') && strncmp(argv[1], "min", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " min\"", NULL);
	    return TCL_ERROR;
	}
	sprintf(interp->result, "%.6f", faPtr->Min());
	return TCL_OK;
    } else if( (c=='m') && strncmp(argv[1], "max", length)==0)  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " max\"", NULL);
	    return TCL_ERROR;
	}
	sprintf(interp->result, "%.6f", faPtr->Max());
	return TCL_OK;
    } else if( (c=='s') && strncmp(argv[1], "scale", length)==0)  {
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " scale <val>\"", NULL);
	    return TCL_ERROR;
	}
	double val;
	sscanf(argv[2], "%lf", &val);
	faPtr->Scale(val);
	return TCL_OK;
    } else if( (c=='o') && strncmp(argv[1], "offset", length)==0)  {
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " offset <val>\"", NULL);
	    return TCL_ERROR;
	}
	double val;
	sscanf(argv[2], "%lf", &val);
	faPtr->Offset(val);
	return TCL_OK;
    } else if( (c=='r') && strncmp(argv[1], "resize", length)==0)  {
	if(argc!=4)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " resize <rows> <cols>\"", NULL);
	    return TCL_ERROR;
	}
	int rows, cols;
	if(Tcl_GetInt(interp, argv[2], &rows) != TCL_OK
	    || Tcl_GetInt(interp, argv[3], &cols) != TCL_OK)  {
	    return TCL_ERROR;
	}
	faPtr->Resize(rows, cols);
	return TCL_OK;
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1], 
		"\":  configure/load/save/clear/transpose/size/val/copy/print/normrows/normcols/norm/min/max/scale/offset/resize",
			 (char *) NULL);
	result = TCL_ERROR;
    }
    return result;
}

int farrayConfigure(floatArray *faPtr, Tcl_Interp *interp, 
		    char *cmdName, int argc, char **argv)
{	/* Handle arguments specifically to configure */
    int i = 0;
    char c, *s;
    int len;

    while(++i < argc)  {
	if(argv[i][0] != '-')  {
	    Tcl_AppendResult(interp, "configure error: should be \"", 
			     cmdName, " config {-mode} ?val?", NULL);
	    return TCL_ERROR;
	}
	s = &(argv[i][1]);  c = *s;  len = strlen(s);
	if( (c=='m') && strncmp(s,"mode",len)==0)  {
	    char mc = 'r';
	    ++i;
	    if(i==argc)  {
		// mc = snd->getMode();
	    } else {
		char *arg = argv[i];
		int larg = strlen(arg);
		char argi = *arg;
		if( (argi=='r') && strncmp(arg, "read", larg)==0)  {
		    // mc = snd->setMode('r');
		} else if( (argi=='w') && strncmp(arg, "write", larg)==0)  {
		    // mc = snd->setMode('w');
		} else {
		    Tcl_AppendResult(interp, "error: should be \"", 
				     cmdName, " mode {read|write}\"", NULL);
		    return TCL_ERROR;
		}
	    }
	    if(mc=='r')
		sprintf(interp->result, "read");
	    else
		sprintf(interp->result, "write");
	} else {
	    Tcl_AppendResult(interp, "unrecognized config option \"", argv[i],
			     NULL);
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

void farrayDelete(ClientData clientData)
{	/* a particular soundObj is being deleted */
    floatArray *faPtr = (floatArray *)clientData;
    delete faPtr;
    /* we really want to unregister the hash table entry here too. */
    int deleted = 0;
    Tcl_HashSearch hs;
    Tcl_HashEntry *hePtr;
    hePtr = Tcl_FirstHashEntry(&farrayTable, &hs);
    while(hePtr)  {
	if(Tcl_GetHashValue(hePtr)==(ClientData)faPtr)  {  // found it
	    Tcl_DeleteHashEntry(hePtr);
	    deleted = 1;
	    hePtr=0;
	} else {
	    hePtr = Tcl_NextHashEntry(&hs);
	}
    }
    if(!deleted)  {
	fprintf(stderr, "farrayDelete: could not dehash 0x%lx\n", faPtr);
    }
}
    
