/* 
 * tkImg.cc 
 *
 *	Show some grayscale data in a pixmap.  Based on tkSquare.c
 * 
 * dpwe@interval.com 1994aug12
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>	/* for log() proto */
#include "tk.h"
#include <assert.h>

/* if LIVE_RECOLOR, effects of scale & offset are applied at redraw, for 
   faster adjustment of brightness/contrast.  If !LIVE_RECOLOR, they 
   are baked-in at rescan time (when reading the data file), meaning the 
   data must be rescanned after scale & offset are changed. */
#define LIVE_RECOLOR 1

/* if COLORMAP, will query environment variables ESPS_BASE and 
   COLORMAP_STEM to build the name of an ESPS-style colormap 
   defining 115 nice colors for showing the image */
#define COLORMAP 1

/* if USEDAT, files are read as DAT files */
/* #define USEDAT 1 */
/* if USEFLOAT, DAT files are read as float data, converted on read */
/* #define READFLOAT 1 */

#ifndef MIN
#define MIN(a,b)  ((a)<(b)?(a):(b))
#define MAX(a,b)  ((a)>(b)?(a):(b))
#endif /* MIN */

/*
 * A data structure of the following type is kept for each image
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the image.  NULL
				 * means window has been deleted but
				 * widget record hasn't been cleaned up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    int borderWidth;		/* Width of 3-D border around whole widget. */
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		/* appearance of whole window */

    int x, y;			/* work area to be used, derived frm win&pad */
    int xsize, ysize;

    char *farrayname;		/* source floatarray tag */
    unsigned char *data;	/* binary image data */
    int  rows, cols;		/* dimensions of <data> */
#ifdef USEDAT
    int	 dset, chan;		/* channel and dataset within DAT file */
    int	 dsets, chans;		/* actual data of DAT file as read */
#endif /* USEDAT */
    int	 transpose, flipx, flipy;	/* display controls */
    double rbase, rlen, cbase, clen;
                                /* range of data actually to display */
    XColor **colors;		/* our colormap */
    int  ncolors;		/* entries in our colormap */

    /*
     * Information used when displaying widget:
     */

    GC gc;			/* Graphics context for copying from
				 * off-screen pixmap onto screen. */
    int doubleBuffer;		/* Non-zero means double-buffer redisplay
				 * with pixmap;  zero means draw straight
				 * onto the display. */
    int updatePending;		/* Non-zero means a call to ImageDisplay
				 * has already been scheduled. */
    Pixmap pm;
    int	   pmw;
    int    pmh;
    float  pmCbase, pmClen;	/* current contents of pixmap */

    /*
     * scaling parameters 
     */
    double	scale;		/* gray values multiplied by this */
    double 	offset;		/* added to gray values */
    int		doLog;		/* convert raw input data to dB before scale */
} Image;

/*
 * Information used for argv parsing.
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"#d9d9d9", Tk_Offset(Image, bgBorder), TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"white", Tk_Offset(Image, bgBorder), TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(Image, borderWidth), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(Image, doubleBuffer), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(Image, relief), 0},
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(Image, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(Image, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(Image, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(Image, height), 0},
    {TK_CONFIG_STRING, "-farrayname", "fArrayName", "FArrayName",
	"", Tk_Offset(Image, farrayname), 0},
#ifdef USEDAT
    {TK_CONFIG_INT, "-dset", "dset", "Dset", 
	"0", Tk_Offset(Image, dset), 0},
    {TK_CONFIG_INT, "-chan", "chan", "Chan", 
	"0", Tk_Offset(Image, chan), 0},
#endif /* USEDAT */
    {TK_CONFIG_INT, "-transpose", "transpose", "Transpose", 
	"0", Tk_Offset(Image, transpose), 0},
    {TK_CONFIG_INT, "-flipx", "flipx", "Flipx", 
	"0", Tk_Offset(Image, flipx), 0},
    {TK_CONFIG_INT, "-flipy", "flipy", "Flipy", 
	"0", Tk_Offset(Image, flipy), 0},
    {TK_CONFIG_INT, "-rows", "rows", "Rows", 
	"0", Tk_Offset(Image, rows), 0},
    {TK_CONFIG_INT, "-cols", "cols", "Cols", 
	"0", Tk_Offset(Image, cols), 0},
    {TK_CONFIG_DOUBLE, "-rbase", "rbase", "Rbase", 
        "0", Tk_Offset(Image, rbase), 0}, 
    {TK_CONFIG_DOUBLE, "-rlen", "rlen", "Rlen", 
        "0", Tk_Offset(Image, rlen), 0}, 
    {TK_CONFIG_DOUBLE, "-cbase", "cbase", "Cbase", 
        "0", Tk_Offset(Image, cbase), 0}, 
    {TK_CONFIG_DOUBLE, "-clen", "clen", "Clen", 
        "0", Tk_Offset(Image, clen), 0}, 
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale", 
        "1", Tk_Offset(Image, scale), 0}, 
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset", 
        "0", Tk_Offset(Image, offset), 0}, 
    {TK_CONFIG_INT, "-log", "log", "Log", 
	"0", Tk_Offset(Image, doLog), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static int		ImageConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    Image *imagePtr, int argc, char **argv,
			    int flags));
static void		ImageDestroy _ANSI_ARGS_((ClientData clientData));
static void		ImageDisplay _ANSI_ARGS_((ClientData clientData));
static void		ImageEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		ImageWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));
static int 		ImageReload _ANSI_ARGS_((Tcl_Interp *interp, 
						 Image *imagePtr));
extern "C" {
int ImageCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int Image_Init(Tcl_Interp *interp);
};

int Image_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "image", ImageCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

static void ImageGetWorkArea(Image *imagePtr)
{   /* calculate the top left and size of the actual area to draw */
    Tk_Window tkwin = imagePtr->tkwin;

    imagePtr->x = imagePtr->padX;
    imagePtr->y = imagePtr->padY;
    imagePtr->xsize = Tk_Width(tkwin) - 2*imagePtr->padX;
    imagePtr->ysize = Tk_Height(tkwin) - 2*imagePtr->padY;
}

/*
 *--------------------------------------------------------------
 *
 * ImageCmd --
 *
 *	This procedure is invoked to process the "image" Tcl
 *	command.  It creates a new "image" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int
ImageCmd(ClientData clientData,	/* Main window associated with interpreter. */
	 Tcl_Interp *interp, 	/* Current interpreter. */
	 int argc, 		/* Number of arguments. */
	 char **argv)		/* Argument strings. */
{
    Tk_Window main = (Tk_Window) clientData;
    Image *imagePtr;
    Tk_Window tkwin;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Image");

    /*
     * Allocate and initialize the widget record.
     */

    imagePtr = (Image *) ckalloc(sizeof(Image));
    imagePtr->tkwin = tkwin;
    imagePtr->display = Tk_Display(tkwin);
    imagePtr->interp = interp;
    imagePtr->x = 0;
    imagePtr->y = 0;
    imagePtr->xsize = 0;
    imagePtr->ysize = 0;
    imagePtr->padX = 0;
    imagePtr->padY = 0;
    
    imagePtr->borderWidth = 0;
    imagePtr->bgBorder = NULL;
    imagePtr->relief = TK_RELIEF_FLAT;
    imagePtr->gc = None;
    imagePtr->doubleBuffer = 1;
    imagePtr->updatePending = 0;

    imagePtr->farrayname = NULL;
    imagePtr->data     = NULL;
    imagePtr->rows     = 0;
    imagePtr->cols     = 0;
    imagePtr->colors   = NULL;
    imagePtr->ncolors  = 0;
#ifdef USEDAT
    imagePtr->chan = 0;
    imagePtr->dset = 0;
    imagePtr->chans = 0;
    imagePtr->dsets = 0;
#endif /* USEDAT */

    imagePtr->pm  = None;
    imagePtr->pmh = 0;
    imagePtr->pmw = 0;
    imagePtr->pmCbase = 0;
    imagePtr->pmClen  = 0;

    Tk_CreateEventHandler(imagePtr->tkwin, ExposureMask|StructureNotifyMask,
	    ImageEventProc, (ClientData) imagePtr);
    Tcl_CreateCommand(interp, Tk_PathName(imagePtr->tkwin), ImageWidgetCmd,
	    (ClientData) imagePtr, (void (*)(ClientData)) NULL);
    if (ImageConfigure(interp, imagePtr, argc-2, argv+2, 0) != TCL_OK) {
	Tk_DestroyWindow(imagePtr->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(imagePtr->tkwin);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ImageWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
ImageWidgetCmd(ClientData clientData, 	/* Information about image widget. */
	       Tcl_Interp *interp, 	/* Current interpreter. */
	       int argc, 		/* Number of arguments. */
	       char **argv)		/* Argument strings. */
{
    Image *imagePtr = (Image *) clientData;
    int result = TCL_OK;
    int length;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) imagePtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, imagePtr->tkwin, configSpecs,
		    (char *) imagePtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, imagePtr->tkwin, configSpecs,
		    (char *) imagePtr, argv[2], 0);
	} else {
	    result = ImageConfigure(interp, imagePtr, argc-2, argv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, imagePtr->tkwin, configSpecs,
				   (char *) imagePtr, argv[2], 0);
    } else if ((c == 'r') && (strncmp(argv[1], "redraw", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " redraw\"", (char *)NULL);
	    return TCL_ERROR;
	}
	if(imagePtr->pm != None)  {
	    XFreePixmap(Tk_Display(imagePtr->tkwin), imagePtr->pm);
	    imagePtr->pm = None;
	}
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if ((c == 'r') && (strncmp(argv[1], "rescan", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " rescan\"", (char *)NULL);
	    return TCL_ERROR;
	}
	ImageReload(interp, imagePtr);

	if(imagePtr->pm != None)  {
	    XFreePixmap(Tk_Display(imagePtr->tkwin), imagePtr->pm);
	    imagePtr->pm = None;
	}
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if ((c == 'f') && (strncmp(argv[1], "fileinfo", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " fileinfo\"", (char *)NULL);
	    return TCL_ERROR;
	}
#ifdef USEDAT
	sprintf(interp->result, "%d %d %d %d", 
		imagePtr->dsets, imagePtr->chans, 
		imagePtr->rows, imagePtr->cols);
#else /* !USEDAT */
	sprintf(interp->result, "%d %d", 
		imagePtr->rows, imagePtr->cols);
#endif /* USEDAT */
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, cget, redraw, rescan, fileinfo", (char *) NULL);
	goto error;
    }
    if (!imagePtr->updatePending) {
	Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	imagePtr->updatePending = 1;
    }
    Tk_Release((ClientData) imagePtr);
    return result;

    error:
    Tk_Release((ClientData) imagePtr);
    return TCL_ERROR;
}


unsigned char *transposeByteArray(unsigned char *data, int r, int c) 
{   /* transpose an array into a new data space */
    unsigned char *oldd = data;
    unsigned char *newd = (unsigned char *)ckalloc(r*c);
    int i, j;

    /* arrays have cols as fastest varying dimension i.e.
       ROW 0 (col 0, col 1, ... col n-1) ROW 1 (col 0, .. */
    for(j = 0; j<r; ++j)  {
	for(i = 0; i<c; ++i)  {
	    newd[i*r + j] = *oldd++;  /* oldd[ j*cols + i ] */
	}
    }
    ckfree(data);
    return newd;
}

unsigned char *flipxByteArray(unsigned char *data, int r, int c) 
{   /* reverse the order of each row */
    int x,y,xt=c,yt=r;
    unsigned char t;

    for(y = 0; y<yt; ++y)  {
	for(x=0; x<(xt/2); ++x)  {
	    t = data[y*xt+x];
	    data[y*xt+x] = data[y*xt+xt-1-x];
	    data[y*xt+xt-1-x] = t;
	}
    }
    return data;
}

unsigned char *flipyByteArray(unsigned char *data, int r, int c) 
{   /* swap rows */
    int x,y,xt=c,yt=r,ym;
    unsigned char t;

    for(x=0; x<xt; ++x)  {
	for(y = 0; y<yt/2; ++y)  {
	    ym = yt - 1 - y;
	    t = data[x+ y*xt];
	    data[x + y*xt] = data[x + ym*xt];
	    data[x + ym*xt] = t;
	}
    }	  
    return data;
}

unsigned char *loadHexFile(char *filename, int *rows, int *cols)
{   /* load up a hex-ascii file with a header specifying rows and cols */
    int r, c, b, s=1, v=0;
    FILE *in;
    char buf[256];
    unsigned char *data, *d2;

    if( (in = fopen(filename, "r")) == NULL)  {
	fprintf(stderr, "loadHexFile: could not open '%s'\n", filename);
	return NULL;	/* could not open the file */
    }
    /* first line is of form "cc rr 8" to specify rows and cols */
    fgets(buf, 256, in);
    sscanf(buf, "%d %d %d", &r, &c, &b);
    if(b != 8) {
	fprintf(stderr, "loadHexFile: bits of %d is not 8\n", b);
	return NULL;
    }
    /* now we're looking good */
    data = (unsigned char *)ckalloc(r*c);
    d2 = data;
    s = 16;
    while(d2 < (data+r*c))  {
	b = fgetc(in);
	if(b == EOF)  break;
	if(b>='a')	b -= 'a'-'A';
	if((b >= '0' && b <='9') || (b>='A' && b <='F'))  {
	    if(b>='A')  b -= 'A'-('9'+1);
	    b -= '0';	/* b is now 0..15 */
	    assert( 0 <= b && b<=15);
	    v += b*s;
	    if(s == 16)	s = 1;
	    else {
		*d2++ = v;
		v = 0;
		s = 16;
	    }
	}
    }

    *rows = r;
    *cols = c;
    fprintf(stderr, "Read %dx%d (%d) bytes from %s\n", r, c, 
	    d2-data, filename);
    return data;
}

#ifdef USEDAT

extern "C" {
#include <dat.h>
};

unsigned char *loadDATFile(char *filename, int dset, int chan, 
			   int *pdsets, int *pchans, int *rows, int *cols, 
			   double scale, double offset, int dolog)
{   /* load up the image from a DAT file */
    DATFILE *df;
    long dsets,chans,dims[2];
    int tdims;
    int rs, cs, r,c;
    float *fdata, *fp, fd;
    int  id;
    unsigned char *data, *ucp;
    struct d_addr  where;

    if( (df = d_open(filename, "r")) == NULL)  {
	fprintf(stderr, "loadDATFile: could not open '%s'\n", filename);
	return NULL;	/* could not open the file */
    }
    if((tdims=d_getdim(df, &dsets, &chans, dims, 2))!=2)  {
	fprintf(stderr, 
		"loadDATFile: warning: file has %d dimensions; last 2 used\n", 
		tdims);
    }
    if(dsets <= 0 || chans <= 0 || dims[0] <= 0 || dims[1] <= 0)  {
	fprintf(stderr, 
		"loadDATFile: no data in '%s' (%d dsets, %d chans, %d x %d)\n",
		dsets, chans, dims[0], dims[1]);
	d_close(df);
	return NULL;
    }
    if(dset >= dsets)  {
	fprintf(stderr, 
		"loadDATFile: no such dset %d, using %d\n", dset, dsets-1);
	dset = dsets-1;
    }
    if(chan >= chans)  {
	fprintf(stderr, 
		"loadDATFile: no such chan %d, using %d\n", chan, chans-1);
	chan = chans-1;
    }

    rs = dims[0]; cs = dims[1];	/* cols are fastest dimension */
    /* now we're looking good */
    data = (unsigned char *)ckalloc(rs*cs);
#ifdef READFLOAT
    fdata = (float *)ckalloc((sizeof(float))*rs*cs);
#endif /* READFLOAT */

    where.dset = dset;
    where.chan = chan;
    where.dim[0] = 0;
    where.dim[1] = 0;
    d_seek(df, &where, 0);	/* SEEK_SET */
#ifdef READFLOAT
    d_readc(df, (char *)fdata, rs*cs, 'f');	/* read as floats */
#else /* !READFLOAT - READBYTE */
    d_readc(df, (char *)data, rs*cs, 'b');	/* read as bytes */
#endif /* READFLOAT */
    d_close(df);

    *rows = rs;
    *cols = cs;
    *pdsets = dsets;
    *pchans = chans;
/*    fprintf(stderr, "Read DAT %dx%d (%d) bytes from %s\n", rs, cs, rs*cs,  
	    filename);	 */

#ifdef READFLOAT
    /* convert float to bytes as specified */
    ucp = data;
    fp  = fdata;
    for(r=0; r<rs; ++r)  {
	for(c=0; c<cs; ++c)  {
	    if(dolog) {
		fd = 8.686*log(*fp++);	/* convert to dB */
		/* scale and offset applied to dB domain */
		id = 255.0*(scale*fd+offset);
	    } else {			/* linear, map 0..1 to 0..255 */
		fd = (scale**fp++)+offset;
		id = 255.0*fd;
	    }
	    *ucp++ = (id>0)?((id<255)?id:255):0;
	}
    }
    free(fdata);	/* unless we can figure where to keep it */
#endif /* READFLOAT */
    return data;
}    

#endif /* USEDAT */

#include "floatArray.h"
extern Tcl_HashTable farrayTable;

unsigned char *loadFloatArray(char *name, int *rowsPtr, int *colsPtr, 
			      double scale, double offset, int dolog)
{   /* load up the image from a floatArray object */
    int r, c, rs, cs;
    unsigned char *data;
    floatArray *srcFaPtr;
    Tcl_HashEntry *entryPtr;

    entryPtr = Tcl_FindHashEntry(&farrayTable, name);
    if(entryPtr == NULL) {  // no farray registered to that name
	// Tcl_AppendResult(interp, "no such farray: ", name, NULL);
	return NULL;
    }
    srcFaPtr = (floatArray *)Tcl_GetHashValue(entryPtr);
    /* ARGH!  tkImg rotates *cols* fastest, floatArray rotates *rows* ->
       implicit transposition */
    *rowsPtr = rs = srcFaPtr->Cols();
    *colsPtr = cs = srcFaPtr->Rows();
    /* convert float to bytes as specified */
    data = (unsigned char *)ckalloc(rs*cs);
    unsigned char *ucp = data;
    int id;
    float fd, *fp= &(srcFaPtr->El(0,0));

    for(r=0; r<rs; ++r)  {
	for(c=0; c<cs; ++c)  {
	    if(dolog) {
		fd = 8.686*log(*fp++);	/* convert to dB */
		/* scale and offset applied to dB domain */
		id = 255.0*(scale*fd+offset);
	    } else {			/* linear, map 0..1 to 0..255 */
		fd = (scale**fp++)+offset;
		id = 255.0*fd;
	    }
	    *ucp++ = (id>0)?((id<255)?id:255):0;
	}
    }
    return data;
}

XColor **initColors(int *pncolors, Tcl_Interp *interp, Tk_Window tkwin)
{	/* grab some colors */
    int ncolors = 32;
    int i;
    char name[32];
    XColor **xcarray;
    int r,g,b;
    Tk_Uid nameId;
#ifdef COLORMAP
    char *cmapStem = "/lib/waves/colormaps/TImap";
    char cmapFname[128];
    char *s;
    FILE *cmapFile = NULL;

    if( (s = getenv("ESPS_BASE")) != NULL) {
	strcpy(cmapFname, s);
	if( (s = getenv("COLORMAP_STEM")) != NULL) {
	    strcat(cmapFname, s);
	} else {
	    strcat(cmapFname, cmapStem);
	}
	if( (cmapFile = fopen(cmapFname, "r")) != NULL) {
	    ncolors = 115;  	/* standard size of ESPS colormap */
	}
    }
#endif /* COLORMAP */
    xcarray = (XColor **)ckalloc(ncolors * sizeof(XColor *));
    for(i = 0; i<ncolors; ++i)  {
	r = i*255/(ncolors-1);
	g = i*255/(ncolors-1);
	b = i*255/(ncolors-1);
#ifdef COLORMAP
	if(cmapFile) {
	    fscanf(cmapFile, "%d %d %d", &r, &g, &b);
	}
#endif /* COLORMAP */
	sprintf(name, "#%02x%02x%02x", r,g,b);
	nameId = Tk_GetUid(name);
	xcarray[i] = Tk_GetColor(interp, tkwin, nameId);
	if(xcarray[i] == NULL)  {
	    ncolors = i;
	}
    }
/*    fprintf(stderr, "initColors: made %d\n", ncolors);   */
#ifdef COLORMAP
    if(cmapFile) {
	fclose(cmapFile);
    }
#endif /* COLORMAP */
    *pncolors = ncolors;
    return xcarray;
}

static int ImageReload(Tcl_Interp *interp, 	/* Used for error reporting. */
		       Image *imagePtr)		/* Information about widget. */
{   /* reread the datafile into the internal buffer */
    int rc = 1;
    double scale = imagePtr->scale;
    double offset = imagePtr->offset;

#ifdef LIVE_RECOLOR
    scale = 1.0;
    offset = 0.0;
#endif /* LIVE_RECOLOR */

    if(imagePtr->data)	ckfree((char *)imagePtr->data);
    if(imagePtr->farrayname)  {
	int tmp;
#ifdef USEDAT
	imagePtr->data = loadDATFile(imagePtr->farrayname, 
				     imagePtr->dset, imagePtr->chan,
				     &(imagePtr->dsets),&(imagePtr->chans),
				     &(imagePtr->rows), &(imagePtr->cols), 
				     scale, offset, imagePtr->doLog);
#else /* !USEDAT */
/*	imagePtr->data = loadHexFile(imagePtr->farrayname,
				     &(imagePtr->rows), &(imagePtr->cols));
 */
	imagePtr->data = loadFloatArray(imagePtr->farrayname,
					&(imagePtr->rows), &(imagePtr->cols), 
					imagePtr->scale, imagePtr->offset, 
					imagePtr->doLog);
#endif /* USEDAT */
	if(!imagePtr->transpose)  {
	    imagePtr->data = transposeByteArray(imagePtr->data, 
						imagePtr->rows,
						imagePtr->cols);
	    tmp = imagePtr->rows; imagePtr->rows = imagePtr->cols;
	    imagePtr->cols = tmp;
	}
	if(imagePtr->flipx) 
	    imagePtr->data = flipxByteArray(imagePtr->data, 
					    imagePtr->rows,
					    imagePtr->cols);
	if(imagePtr->flipy) 
	    imagePtr->data = flipyByteArray(imagePtr->data, 
					    imagePtr->rows,
					    imagePtr->cols);
    }
    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * ImageConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a image widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for imagePtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ImageConfigure(Tcl_Interp *interp,	/* Used for error reporting. */
	       Image *imagePtr, 	/* Information about widget. */
	       int argc, 		/* Number of valid entries in argv. */
	       char **argv, 		/* Arguments. */
	       int flags)		/* Flags to pass to
					 * Tk_ConfigureWidget. */
{
    int needNewData = (imagePtr->data==NULL)?1:0;

    if (Tk_ConfigureWidget(interp, imagePtr->tkwin, configSpecs,
	    argc, argv, (char *) imagePtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Set the background for the window and create a graphics context
     * for use during redisplay.
     */

    Tk_SetWindowBackground(imagePtr->tkwin,
	    Tk_3DBorderColor(imagePtr->bgBorder)->pixel);
    if ((imagePtr->gc == None) && (imagePtr->doubleBuffer)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	imagePtr->gc = Tk_GetGC(imagePtr->tkwin,
		GCFunction|GCGraphicsExposures, &gcValues);
    }

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */

    Tk_GeometryRequest(imagePtr->tkwin, imagePtr->width, imagePtr->height);
    Tk_SetInternalBorder(imagePtr->tkwin, imagePtr->borderWidth);

    if(imagePtr->colors == NULL) {
	imagePtr->colors = initColors(&(imagePtr->ncolors), 
				      imagePtr->interp, imagePtr->tkwin);
    }

/*  return (imagePtr->colors != NULL && imagePtr->data != NULL)?TCL_OK:TCL_ERR;
 */

    if(needNewData)  {
	ImageReload(interp, imagePtr);
    }

    if (!imagePtr->updatePending) {
	Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	imagePtr->updatePending = 1;
    }

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ImageEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on images.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
ImageEventProc(ClientData clientData,	/* Information about window. */
	       XEvent *eventPtr)	/* Information about event. */
{
    Image *imagePtr = (Image *) clientData;

    if (eventPtr->type == Expose) {
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if (eventPtr->type == ConfigureNotify) {
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if (eventPtr->type == DestroyNotify) {
	Tcl_DeleteCommand(imagePtr->interp, Tk_PathName(imagePtr->tkwin));
	imagePtr->tkwin = NULL;
	if (imagePtr->updatePending) {
	    Tk_CancelIdleCall(ImageDisplay, (ClientData) imagePtr);
	}
	Tk_EventuallyFree((ClientData) imagePtr, (Tcl_FreeProc*)ImageDestroy);
    }
}

/********* cacheColor (cchClr) routines to reduce XSetColor ohead **********/

#define CCH_BUF_SIZE	256	/* buffer this many instances of each color */
typedef struct {	/* record of rectangles to set to a given color */
    int n;		/* how many entries in the rects array */
    XRectangle rects[CCH_BUF_SIZE];	/* recorded rectangles */
} CCH_ENT;

/* static parameters set by initialize, used later */
static CCH_ENT *cchClrRoot = NULL;
static XColor **cchClrCmap = NULL;
static int      cchClrNclr = 0;
static Display *cchClrDisp = None;
static Drawable cchClrD    = None;
static GC       cchClrGC   = None;

static void cchClrInit(Display *disp, Drawable d, GC gc, 
		       XColor **colormap, int ncolors)
{   /* Initializes a sequence of calls to cchClrColor and holds the data 
       actually needed to draw the colors, when the time comes */
    int i;
    cchClrDisp = disp;
    cchClrD    = d;
    cchClrGC   = gc;
    cchClrCmap = colormap;
    if(ncolors!=cchClrNclr)  {	/* do we need a new space for recording? */
	cchClrNclr = ncolors;
	if(cchClrRoot)
	    ckfree((char *)cchClrRoot);
	cchClrRoot = (CCH_ENT *)ckalloc(sizeof(CCH_ENT)*ncolors);
    }
    for(i=0; i<ncolors; ++i)
	cchClrRoot[i].n = 0;	/* make sure we know it's all empty */
}

static void cchClrDoColor(int color)
{	/* actually draw all the rects for a given color */
    int n;

    if( (n = cchClrRoot[color].n) > 0)  {
	XSetForeground(cchClrDisp, cchClrGC, 
		       cchClrCmap[color]->pixel);
	XFillRectangles(cchClrDisp, cchClrD, cchClrGC, 
			cchClrRoot[color].rects, n);
	cchClrRoot[color].n = 0;	/* these rects all finished with */
    }
}

static void cchClrColor(int x, int y, int w, int h, int color)
{	/* buffer up a rectangle for a particular color */
    int n;
    XRectangle *pr;

#ifdef LIVE_RECOLOR
    if (color < 0) color = 0;
    if (color >= cchClrNclr) color = cchClrNclr-1;
#else /* !LIVE_RECOLOR */
    assert(color < cchClrNclr);
#endif /* LIVE_RECOLOR */

    n = cchClrRoot[color].n;
    pr = &(cchClrRoot[color].rects[n]);
    pr->x = x; pr->y = y; pr->width = w; pr->height = h;
    if((cchClrRoot[color].n = ++n) == CCH_BUF_SIZE) {
	/* we've filled the buffer for this color, so empty it now */
	cchClrDoColor(color);	/* (which clears n as a side effect) */
    }
}

static void cchClrFinish(void)
{ 	/* After all the calls to cchClrColor are done, go through and 
	   make all the X calls */
    int i;

    for(i=0; i<cchClrNclr; ++i)
	cchClrDoColor(i);
}

/***************** end of cchClr stuff **********************/


/*
 *--------------------------------------------------------------
 *
 * ImageDisplay --
 *
 *	This procedure redraws the contents of a image window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
ImageDisplay(ClientData clientData)	/* Information about window. */
{
    Image *imagePtr = (Image *) clientData;
    Tk_Window tkwin = imagePtr->tkwin;
    Drawable d;
    int w,h,v;
    double scale = imagePtr->scale;
    double offset = imagePtr->offset;

    imagePtr->updatePending = 0;
    if (!Tk_IsMapped(tkwin)) {
	return;
    }

    /* Create a pixmap for double-buffering, if necessary. */
    w = Tk_Width(tkwin); h = Tk_Height(tkwin);
    if (imagePtr->doubleBuffer) {
	if(imagePtr->pm == None || imagePtr->pmw != w || imagePtr->pmh != h) {
	    if(imagePtr->pm != None)
		XFreePixmap(Tk_Display(tkwin), imagePtr->pm);
	    imagePtr->pm = XCreatePixmap(Tk_Display(tkwin), Tk_WindowId(tkwin),
					 w, h, 
				     DefaultDepthOfScreen(Tk_Screen(tkwin)));
	    imagePtr->pmw = w;
	    imagePtr->pmh = h;
	    imagePtr->pmCbase = 0;
	    imagePtr->pmClen  = 0;
	    /* Redraw the widget's background and border only once per pm. */
	    Tk_Fill3DRectangle(tkwin, imagePtr->pm, 
			       imagePtr->bgBorder,
			       0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
			       imagePtr->borderWidth, imagePtr->relief);

	}
	d = imagePtr->pm;
    } else {
	d = Tk_WindowId(tkwin);
	/* Redraw the widget's background and border. */
	Tk_Fill3DRectangle(tkwin, d, imagePtr->bgBorder,
			   0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
			   imagePtr->borderWidth, imagePtr->relief);
    }

    /* Display the image. */
    /* first, set up work area ptr (->x, ->xsize etc) */
    ImageGetWorkArea(imagePtr);
    if(imagePtr->data == NULL)  {
	Tk_Fill3DRectangle(tkwin, d, imagePtr->bgBorder,
			   imagePtr->x, imagePtr->y, 
			   imagePtr->xsize, imagePtr->ysize,
			   imagePtr->borderWidth, TK_RELIEF_SUNKEN);
    } else {
	/* problem now is to fit */
	int x1, x2, y1, y2;
	int i,j;
	int arxmin, arxmax, arymin, arymax;
	float arw, arh;

	/* one Fill command for each data array element */
	arxmin = (int)imagePtr->cbase;
        if( (arw = imagePtr->clen) == 0) {
	    arw = imagePtr->cols - imagePtr->cbase;
	}
	arxmax = (int)(0.99+imagePtr->cbase+arw);
	/* does the current pixmap contain reusable data? */
	if(imagePtr->doubleBuffer && imagePtr->pmClen == arw) {
	    float cbase = imagePtr->cbase;
	    float pmcb  = imagePtr->pmCbase;
	    int	sx, sy=imagePtr->y, dx, dy=imagePtr->y;
	    unsigned int sw, sh=imagePtr->ysize;
	    
	    if(pmcb >= cbase && pmcb < cbase+arw) {
		/* can shift data right */
		sx = imagePtr->x;
		dx = imagePtr->x+(.5+((float)imagePtr->xsize)*((pmcb-cbase)/arw));
		sw = imagePtr->x+imagePtr->xsize - dx;
		XCopyArea(imagePtr->display, d, d, imagePtr->gc, 
			  sx, sy, sw, sh, dx, dy);
	/*	fprintf(stderr, "sx %d, sy %d, dx %d, dy %d, w %d h %d\n",
			sx,sy,dx,dy,w,h);
	 */
		arxmax = (int)(0.99+pmcb);
	    } else if(cbase >= pmcb && cbase < pmcb+arw) {
		/* can shift pixels left */
		sx = imagePtr->x+(.5+((float)imagePtr->xsize)*((cbase-pmcb)/arw));
		dx = imagePtr->x;
		sw = imagePtr->x+imagePtr->xsize - sx;
		XCopyArea(imagePtr->display, d, d, imagePtr->gc, 
			  sx, sy, sw, sh, dx, dy);
	/*	fprintf(stderr, "sx %d, sy %d, dx %d, dy %d, w %d h %d\n",
			sx,sy,dx,dy,sw,sh);
	 */
		arxmin = (int)(pmcb+arw);
	    }
	}
	if(arxmin < 0)  {
	    int x = imagePtr->x;
	    unsigned int w = imagePtr->xsize*((-imagePtr->cbase)/arw);
	    int	y=imagePtr->y, h=imagePtr->ysize;
/*
	    XClearArea(imagePtr->display,d,x,y,w,h,0);
 */
	    XClearArea(imagePtr->display,Tk_WindowId(tkwin),x,y,w,h,0);
	    if(imagePtr->doubleBuffer)  {
		XCopyArea(imagePtr->display, Tk_WindowId(tkwin), d, 
			  imagePtr->gc, 
			  x, y, w, h, x, y);
	    }
	    arxmin = 0;
	}
	if(arxmax > imagePtr->cols)  {
	    int x = imagePtr->x 
		    + imagePtr->xsize*((imagePtr->cols-imagePtr->cbase)/arw);
	    unsigned int w = imagePtr->xsize - (x-imagePtr->x);
	    int	y=imagePtr->y, h=imagePtr->ysize;
	    XClearArea(imagePtr->display,Tk_WindowId(tkwin),x,y,w,h,0);
	    if(imagePtr->doubleBuffer)  {
		XCopyArea(imagePtr->display, Tk_WindowId(tkwin), d, 
			  imagePtr->gc, 
			  x, y, w, h, x, y);
	    }
	    arxmax = imagePtr->cols;
	}
	     
/* 	fprintf(stderr, "redrawing cols %d..%d (data %dx%d)\n", 
		arxmin, arxmax,imagePtr->rows, imagePtr->cols);	*/
	arymin = (int)imagePtr->rbase;
        if( (arh = imagePtr->rlen) == 0) {
	    arh = imagePtr->rows - imagePtr->rbase;
	}
	arymax = (int)(0.99+imagePtr->rbase+arh);

	cchClrInit(imagePtr->display, d, imagePtr->gc, 
		   imagePtr->colors, imagePtr->ncolors);
	for(i = arxmin; i<arxmax; ++i)  {
	    for(j = arymin; j<arymax; ++j)  {
		x1 = imagePtr->x 
		    + imagePtr->xsize*(((float)i-imagePtr->cbase)/arw);
		if(x1 < imagePtr->x)	x1 = imagePtr->x;
		x2 = imagePtr->x 
		    + imagePtr->xsize*(((float)i+1.0-imagePtr->cbase)/arw);
		if(x2 > imagePtr->x + imagePtr->xsize)
		    x2 = imagePtr->x + imagePtr->xsize;
		/* y1 is *lower* edge, meaning *larger* val of y */
		y1 = imagePtr->y + imagePtr->ysize 
		    - imagePtr->ysize*(((float)j-imagePtr->rbase)/arh);
		if(y1 > imagePtr->y + imagePtr->ysize)
		    y1 = imagePtr->y + imagePtr->ysize;
		y2 = imagePtr->y  + imagePtr->ysize  
		    - imagePtr->ysize*(((float)j+1.0-imagePtr->rbase)/arh);
		if(y2 < imagePtr->y)	y2 = imagePtr->y;
		/* y2 is supposed to be the smaller of the two */
		if(y2 > y1)	y2 = y1;
		/* data stored with cols most rapid dimension */
		v = imagePtr->data[i + imagePtr->cols*j];
#ifdef LIVE_RECOLOR
		v = (int)((((v-128)/255.0)*scale+offset+0.5)*imagePtr->ncolors);
#else /* !LIVE_RECOLOR */
		v = (v*imagePtr->ncolors)/256;
#endif /* LIVE_RECOLOR */
		cchClrColor(x1, y2, x2-x1, y1-y2, v);
	    }
	}
	cchClrFinish();
	if (imagePtr->doubleBuffer) {
	    imagePtr->pmCbase = imagePtr->cbase;
	    imagePtr->pmClen  = arw /* imagePtr->clen */ ;
	}
    }
    /* If double-buffered, copy to the screen */
    if (imagePtr->doubleBuffer) {
	XCopyArea(Tk_Display(tkwin), imagePtr->pm, Tk_WindowId(tkwin), 
		  imagePtr->gc,
		0, 0, Tk_Width(tkwin), Tk_Height(tkwin), 0, 0);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ImageDestroy --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a image at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the image is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
ImageDestroy(ClientData clientData)	/* Info about image widget. */
{
    Image *imagePtr = (Image *) clientData;
    int i;

    Tk_FreeOptions(configSpecs, (char *) imagePtr, imagePtr->display, 0);
    if (imagePtr->gc != None) {
	Tk_FreeGC(imagePtr->display, imagePtr->gc);
    }
    if(imagePtr->colors) {
	for(i = 0; i<imagePtr->ncolors; ++i)  
	    Tk_FreeColor(imagePtr->colors[i]);
	ckfree((char *)imagePtr->colors);
	imagePtr->colors = NULL;
	imagePtr->ncolors = 0;
    }
    if(imagePtr->data)  {
	ckfree((char *)imagePtr->data);
	imagePtr->data = NULL;
    }
    ckfree((char *) imagePtr);
}
