/* 
 * tkShmImg.cc 
 *
 *	Show some grayscale data.  Based on tkSquare.c
 * 	This version (attempts to) use)s( the MIT Shared Memory extension
 * 1996nov05 dpwe@icsi.berkeley.edu after dpwe@interval.com 1994aug12
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>	/* for log() proto */
#include "tk.h"
#include <assert.h>

/* MIT-SHM includes */
#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/Xutil.h>	/* for XDestroyImage macro */

#define SHM 1	/* invoke SHM code */

#ifndef MIN
#define MIN(a,b)  ((a)<(b)?(a):(b))
#define MAX(a,b)  ((a)>(b)?(a):(b))
#endif /* MIN */

/*
 * A data structure of the following type is kept for each image
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the image.  NULL
				 * means window has been deleted but
				 * widget record hasn't been cleaned up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    int borderWidth;		/* Width of 3-D border around whole widget. */
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		/* appearance of whole window */

    int x, y;			/* work area to be used, derived frm win&pad */
    int xsize, ysize;

    char *farrayname;		/* source floatarray tag */
    unsigned char *data;	/* binary image data */
    int  rows, cols;		/* dimensions of <data> */
#ifdef USEDAT
    int	 dset, chan;		/* channel and dataset within DAT file */
    int	 dsets, chans;		/* actual data of DAT file as read */
#endif /* USEDAT */
    int	 transpose, flipx, flipy;	/* display controls */
    double rbase, rlen, cbase, clen;
                                /* range of data actually to display */
    XColor **colors;		/* our colormap */
    int  ncolors;		/* entries in our colormap */

    /*
     * Information used when displaying widget:
     */

    GC gc;			/* Graphics context for copying from
				 * off-screen pixmap onto screen. */
    int updatePending;		/* Non-zero means a call to ImageDisplay
				 * has already been scheduled. */
    int	   pmw;
    int    pmh;
    float  pmCbase, pmClen;	/* current contents of pixmap */

    /*
     * scaling parameters 
     */
    double	scale;		/* gray values multiplied by this */
    double 	offset;		/* added to gray values */
    int		doLog;		/* convert raw input data to dB before scale */
    /*
     * Shared memory stuff
     */
    XShmSegmentInfo shminfo;
    XImage *image;
    char* actdata;
} ShmImage;

/*
 * Information used for argv parsing.
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"#d9d9d9", Tk_Offset(ShmImage, bgBorder), TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"white", Tk_Offset(ShmImage, bgBorder), TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(ShmImage, borderWidth), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(ShmImage, relief), 0},
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(ShmImage, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(ShmImage, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(ShmImage, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(ShmImage, height), 0},
    {TK_CONFIG_STRING, "-farrayname", "fArrayName", "FArrayName",
	"", Tk_Offset(ShmImage, farrayname), 0},
#ifdef USEDAT
    {TK_CONFIG_INT, "-dset", "dset", "Dset", 
	"0", Tk_Offset(ShmImage, dset), 0},
    {TK_CONFIG_INT, "-chan", "chan", "Chan", 
	"0", Tk_Offset(ShmImage, chan), 0},
#endif /* USEDAT */
    {TK_CONFIG_INT, "-transpose", "transpose", "Transpose", 
	"0", Tk_Offset(ShmImage, transpose), 0},
    {TK_CONFIG_INT, "-flipx", "flipx", "Flipx", 
	"0", Tk_Offset(ShmImage, flipx), 0},
    {TK_CONFIG_INT, "-flipy", "flipy", "Flipy", 
	"0", Tk_Offset(ShmImage, flipy), 0},
    {TK_CONFIG_INT, "-rows", "rows", "Rows", 
	"0", Tk_Offset(ShmImage, rows), 0},
    {TK_CONFIG_INT, "-cols", "cols", "Cols", 
	"0", Tk_Offset(ShmImage, cols), 0},
    {TK_CONFIG_DOUBLE, "-rbase", "rbase", "Rbase", 
        "0", Tk_Offset(ShmImage, rbase), 0}, 
    {TK_CONFIG_DOUBLE, "-rlen", "rlen", "Rlen", 
        "0", Tk_Offset(ShmImage, rlen), 0}, 
    {TK_CONFIG_DOUBLE, "-cbase", "cbase", "Cbase", 
        "0", Tk_Offset(ShmImage, cbase), 0}, 
    {TK_CONFIG_DOUBLE, "-clen", "clen", "Clen", 
        "0", Tk_Offset(ShmImage, clen), 0}, 
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale", 
        "1", Tk_Offset(ShmImage, scale), 0}, 
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset", 
        "0", Tk_Offset(ShmImage, offset), 0}, 
    {TK_CONFIG_INT, "-log", "log", "Log", 
	"0", Tk_Offset(ShmImage, doLog), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static int		ShmImageConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    ShmImage *imagePtr, int argc, char **argv,
			    int flags));
static void		ShmImageDestroy _ANSI_ARGS_((ClientData clientData));
static void		ShmImageDisplay _ANSI_ARGS_((ClientData clientData));
static void		ShmImageEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		ShmImageWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));
static int 		ShmImageReload _ANSI_ARGS_((Tcl_Interp *interp, 
						 ShmImage *imagePtr));
extern "C" {
int ShmImageCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int ShmImage_Init(Tcl_Interp *interp);
};

int ShmImage_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "shmimage", ShmImageCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    int maj, min;
    Bool pixm;
    Display *disp = Tk_Display(main);
    Status s = XShmQueryVersion(disp, &maj, &min, &pixm);
    int pmfmt = XShmPixmapFormat(disp);
    fprintf(stderr, "XShm: stat=%d, ver=%d.%d, pixm=%d, pxfmt=%d\n", s, maj, min, pixm, pmfmt);
    return TCL_OK;
}

static void ShmImageGetWorkArea(ShmImage *imagePtr)
{   /* calculate the top left and size of the actual area to draw */
    Tk_Window tkwin = imagePtr->tkwin;

    imagePtr->x = imagePtr->padX;
    imagePtr->y = imagePtr->padY;
    imagePtr->xsize = Tk_Width(tkwin) - 2*imagePtr->padX;
    imagePtr->ysize = Tk_Height(tkwin) - 2*imagePtr->padY;
}

/*
 *--------------------------------------------------------------
 *
 * ShmImageCmd --
 *
 *	This procedure is invoked to process the "ShmImage" Tcl
 *	command.  It creates a new "ShmImage" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int
ShmImageCmd(ClientData clientData,	/* Main window associated with interpreter. */
	 Tcl_Interp *interp, 	/* Current interpreter. */
	 int argc, 		/* Number of arguments. */
	 char **argv)		/* Argument strings. */
{
    Tk_Window main = (Tk_Window) clientData;
    ShmImage *imagePtr;
    Tk_Window tkwin;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "ShmImage");

    /*
     * Allocate and initialize the widget record.
     */

    imagePtr = (ShmImage *) ckalloc(sizeof(ShmImage));
    imagePtr->tkwin = tkwin;
    imagePtr->display = Tk_Display(tkwin);
    imagePtr->interp = interp;
    imagePtr->x = 0;
    imagePtr->y = 0;
    imagePtr->xsize = 0;
    imagePtr->ysize = 0;
    imagePtr->padX = 0;
    imagePtr->padY = 0;
    
    imagePtr->borderWidth = 0;
    imagePtr->bgBorder = NULL;
    imagePtr->relief = TK_RELIEF_FLAT;
    imagePtr->gc = None;
    imagePtr->updatePending = 0;

    imagePtr->farrayname = NULL;
    imagePtr->data     = NULL;
    imagePtr->rows     = 0;
    imagePtr->cols     = 0;
    imagePtr->colors   = NULL;
    imagePtr->ncolors  = 0;
#ifdef USEDAT
    imagePtr->chan = 0;
    imagePtr->dset = 0;
    imagePtr->chans = 0;
    imagePtr->dsets = 0;
#endif /* USEDAT */

    imagePtr->pmh = 0;
    imagePtr->pmw = 0;
    imagePtr->pmCbase = 0;
    imagePtr->pmClen  = 0;

    /* imagePtr->shminfo = 0; */	/* just space, not a pointer */
    imagePtr->image   = 0;

    Tk_CreateEventHandler(imagePtr->tkwin, ExposureMask|StructureNotifyMask,
	    ShmImageEventProc, (ClientData) imagePtr);
    Tcl_CreateCommand(interp, Tk_PathName(imagePtr->tkwin), ShmImageWidgetCmd,
	    (ClientData) imagePtr, (void (*)(ClientData)) NULL);
    if (ShmImageConfigure(interp, imagePtr, argc-2, argv+2, 0) != TCL_OK) {
	Tk_DestroyWindow(imagePtr->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(imagePtr->tkwin);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ShmImageWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
ShmImageWidgetCmd(ClientData clientData, 	/* Information about image widget. */
	       Tcl_Interp *interp, 	/* Current interpreter. */
	       int argc, 		/* Number of arguments. */
	       char **argv)		/* Argument strings. */
{
    ShmImage *imagePtr = (ShmImage *) clientData;
    int result = TCL_OK;
    int length;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) imagePtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, imagePtr->tkwin, configSpecs,
		    (char *) imagePtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, imagePtr->tkwin, configSpecs,
		    (char *) imagePtr, argv[2], 0);
	} else {
	    result = ShmImageConfigure(interp, imagePtr, argc-2, argv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, imagePtr->tkwin, configSpecs,
				   (char *) imagePtr, argv[2], 0);
    } else if ((c == 'r') && (strncmp(argv[1], "redraw", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " redraw\"", (char *)NULL);
	    return TCL_ERROR;
	}
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if ((c == 'r') && (strncmp(argv[1], "rescan", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " rescan\"", (char *)NULL);
	    return TCL_ERROR;
	}
	ShmImageReload(interp, imagePtr);

	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if ((c == 'f') && (strncmp(argv[1], "fileinfo", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " fileinfo\"", (char *)NULL);
	    return TCL_ERROR;
	}
#ifdef USEDAT
	sprintf(interp->result, "%d %d %d %d", 
		imagePtr->dsets, imagePtr->chans, 
		imagePtr->rows, imagePtr->cols);
#else /* !USEDAT */
	sprintf(interp->result, "%d %d", 
		imagePtr->rows, imagePtr->cols);
#endif /* USEDAT */
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, cget, redraw, rescan, fileinfo", (char *) NULL);
	goto error;
    }
    if (!imagePtr->updatePending) {
	Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	imagePtr->updatePending = 1;
    }
    Tk_Release((ClientData) imagePtr);
    return result;

    error:
    Tk_Release((ClientData) imagePtr);
    return TCL_ERROR;
}

static unsigned char *transposeByteArray(unsigned char *data, int r, int c) 
{   /* transpose an array into a new data space */
    unsigned char *oldd = data;
    unsigned char *newd = (unsigned char *)ckalloc(r*c);
    int i, j;

    /* arrays have cols as fastest varying dimension i.e.
       ROW 0 (col 0, col 1, ... col n-1) ROW 1 (col 0, .. */
    for(j = 0; j<r; ++j)  {
	for(i = 0; i<c; ++i)  {
	    newd[i*r + j] = *oldd++;  /* oldd[ j*cols + i ] */
	}
    }
    ckfree(data);
    return newd;
}

static unsigned char *flipxByteArray(unsigned char *data, int r, int c) 
{   /* reverse the order of each row */
    int x,y,xt=c,yt=r;
    unsigned char t;

    for(y = 0; y<yt; ++y)  {
	for(x=0; x<(xt/2); ++x)  {
	    t = data[y*xt+x];
	    data[y*xt+x] = data[y*xt+xt-1-x];
	    data[y*xt+xt-1-x] = t;
	}
    }
    return data;
}

static unsigned char *flipyByteArray(unsigned char *data, int r, int c) 
{   /* swap rows */
    int x,y,xt=c,yt=r,ym;
    unsigned char t;

    for(x=0; x<xt; ++x)  {
	for(y = 0; y<yt/2; ++y)  {
	    ym = yt - 1 - y;
	    t = data[x+ y*xt];
	    data[x + y*xt] = data[x + ym*xt];
	    data[x + ym*xt] = t;
	}
    }	  
    return data;
}

static unsigned char *loadHexFile(char *filename, int *rows, int *cols)
{   /* load up a hex-ascii file with a header specifying rows and cols */
    int r, c, b, s=1, v=0;
    FILE *in;
    char buf[256];
    unsigned char *data, *d2;

    if( (in = fopen(filename, "r")) == NULL)  {
	fprintf(stderr, "loadHexFile: could not open '%s'\n", filename);
	return NULL;	/* could not open the file */
    }
    /* first line is of form "cc rr 8" to specify rows and cols */
    fgets(buf, 256, in);
    sscanf(buf, "%d %d %d", &r, &c, &b);
    if(b != 8) {
	fprintf(stderr, "loadHexFile: bits of %d is not 8\n", b);
	return NULL;
    }
    /* now we're looking good */
    data = (unsigned char *)ckalloc(r*c);
    d2 = data;
    s = 16;
    while(d2 < (data+r*c))  {
	b = fgetc(in);
	if(b == EOF)  break;
	if(b>='a')	b -= 'a'-'A';
	if((b >= '0' && b <='9') || (b>='A' && b <='F'))  {
	    if(b>='A')  b -= 'A'-('9'+1);
	    b -= '0';	/* b is now 0..15 */
	    assert( 0 <= b && b<=15);
	    v += b*s;
	    if(s == 16)	s = 1;
	    else {
		*d2++ = v;
		v = 0;
		s = 16;
	    }
	}
    }

    *rows = r;
    *cols = c;
    fprintf(stderr, "Read %dx%d (%d) bytes from %s\n", r, c, 
	    d2-data, filename);
    return data;
}

#ifdef USEDAT

extern "C" {
#include <dat.h>
};

static unsigned char *loadDATFile(char *filename, int dset, int chan, 
			       int *pdsets, int *pchans, int *rows, int *cols, 
				  double scale, double offset, int dolog)
{   /* load up the image from a DAT file */
    DATFILE *df;
    long dsets,chans,dims[2];
    int tdims;
    int rs, cs, r,c;
    float *fdata, *fp, fd;
    int  id;
    unsigned char *data, *ucp;
    struct d_addr  where;

    if( (df = d_open(filename, "r")) == NULL)  {
	fprintf(stderr, "loadDATFile: could not open '%s'\n", filename);
	return NULL;	/* could not open the file */
    }
    if((tdims=d_getdim(df, &dsets, &chans, dims, 2))!=2)  {
	fprintf(stderr, 
		"loadDATFile: warning: file has %d dimensions; last 2 used\n", 
		tdims);
    }
    if(dsets <= 0 || chans <= 0 || dims[0] <= 0 || dims[1] <= 0)  {
	fprintf(stderr, 
		"loadDATFile: no data in '%s' (%d dsets, %d chans, %d x %d)\n",
		dsets, chans, dims[0], dims[1]);
	d_close(df);
	return NULL;
    }
    if(dset >= dsets)  {
	fprintf(stderr, 
		"loadDATFile: no such dset %d, using %d\n", dset, dsets-1);
	dset = dsets-1;
    }
    if(chan >= chans)  {
	fprintf(stderr, 
		"loadDATFile: no such chan %d, using %d\n", chan, chans-1);
	chan = chans-1;
    }

    rs = dims[0]; cs = dims[1];	/* cols are fastest dimension */
    /* now we're looking good */
    data = (unsigned char *)ckalloc(rs*cs);
#define READFLOAT
#ifdef READFLOAT
    fdata = (float *)ckalloc((sizeof(float))*rs*cs);
#endif /* READFLOAT */

    where.dset = dset;
    where.chan = chan;
    where.dim[0] = 0;
    where.dim[1] = 0;
    d_seek(df, &where, 0);	/* SEEK_SET */
#ifdef READFLOAT
    d_readc(df, (char *)fdata, rs*cs, 'f');	/* read as floats */
#else /* !READFLOAT - READBYTE */
    d_readc(df, (char *)data, rs*cs, 'b');	/* read as bytes */
#endif /* READFLOAT */
    d_close(df);

    *rows = rs;
    *cols = cs;
    *pdsets = dsets;
    *pchans = chans;
/*    fprintf(stderr, "Read DAT %dx%d (%d) bytes from %s\n", rs, cs, rs*cs,  
	    filename);	 */

#ifdef READFLOAT
    /* convert float to bytes as specified */
    ucp = data;
    fp  = fdata;
    for(r=0; r<rs; ++r)  {
	for(c=0; c<cs; ++c)  {
	    if(dolog) {
		fd = 8.686*log(*fp++);	/* convert to dB */
		/* scale and offset applied to dB domain */
		id = 255.0*(scale*fd+offset);
	    } else {			/* linear, map 0..1 to 0..255 */
		fd = (scale**fp++)+offset;
		id = 255.0*fd;
	    }
	    *ucp++ = (id>0)?((id<255)?id:255):0;
	}
    }
    free(fdata);	/* unless we can figure where to keep it */
#endif /* READFLOAT */
    return data;
}    

#endif /* USEDAT */

#include "floatArray.h"
extern Tcl_HashTable farrayTable;

static unsigned char *loadFloatArray(char *name, int *rowsPtr, int *colsPtr, 
				     double scale, double offset, int dolog)
{   /* load up the image from a floatArray object */
    int r, c, rs, cs;
    unsigned char *data;
    floatArray *srcFaPtr;
    Tcl_HashEntry *entryPtr;

    entryPtr = Tcl_FindHashEntry(&farrayTable, name);
    if(entryPtr == NULL) {  // no farray registered to that name
	// Tcl_AppendResult(interp, "no such farray: ", name, NULL);
	return NULL;
    }
    srcFaPtr = (floatArray *)Tcl_GetHashValue(entryPtr);
    /* ARGH!  tkImg rotates *cols* fastest, floatArray rotates *rows* ->
       implicit transposition */
    *rowsPtr = rs = srcFaPtr->Cols();
    *colsPtr = cs = srcFaPtr->Rows();
    /* convert float to bytes as specified */
    data = (unsigned char *)ckalloc(rs*cs);
    unsigned char *ucp = data;
    int id;
    float fd, *fp= &(srcFaPtr->El(0,0));

    for(r=0; r<rs; ++r)  {
	for(c=0; c<cs; ++c)  {
	    if(dolog) {
		fd = 8.686*log(*fp++);	/* convert to dB */
		/* scale and offset applied to dB domain */
		id = 255.0*(scale*fd+offset);
	    } else {			/* linear, map 0..1 to 0..255 */
		fd = (scale**fp++)+offset;
		id = 255.0*fd;
	    }
	    *ucp++ = (id>0)?((id<255)?id:255):0;
	}
    }
    return data;
}

static XColor **initColors(int *pncolors, Tcl_Interp *interp, Tk_Window tkwin)
{	/* grab some colors */
    int ncolors = 32;
    int i;
    char name[32];
    XColor **xcarray;
    int r,g,b;
    Tk_Uid nameId;

    xcarray = (XColor **)ckalloc(ncolors * sizeof(XColor *));
    for(i = 0; i<ncolors; ++i)  {
	r = i*255/(ncolors-1);
	g = i*255/(ncolors-1);
	b = i*255/(ncolors-1);
	sprintf(name, "#%02x%02x%02x", r,g,b);
	nameId = Tk_GetUid(name);
	xcarray[i] = Tk_GetColor(interp, tkwin, nameId);
	if(xcarray[i] == NULL)  {
	    ncolors = i;
	}
    }
/*    fprintf(stderr, "initColors: made %d\n", ncolors);   */
    *pncolors = ncolors;
    return xcarray;
}

static int ShmImageReload(Tcl_Interp *interp, 	/* Used for error reporting. */
		       ShmImage *imagePtr)		/* Information about widget. */
{   /* reread the datafile into the internal buffer */
    int rc = 1;
    if(imagePtr->data)	ckfree((char *)imagePtr->data);
    if(imagePtr->farrayname)  {
	int tmp;
#ifdef USEDAT
	imagePtr->data = loadDATFile(imagePtr->farrayname, 
				     imagePtr->dset, imagePtr->chan,
				     &(imagePtr->dsets),&(imagePtr->chans),
				     &(imagePtr->rows), &(imagePtr->cols), 
				     imagePtr->scale, imagePtr->offset, 
				     imagePtr->doLog);
#else /* !USEDAT */
/*	imagePtr->data = loadHexFile(imagePtr->farrayname,
				     &(imagePtr->rows), &(imagePtr->cols));
 */
	imagePtr->data = loadFloatArray(imagePtr->farrayname,
					&(imagePtr->rows), &(imagePtr->cols), 
					imagePtr->scale, imagePtr->offset, 
					imagePtr->doLog);
#endif /* USEDAT */
	if(!imagePtr->transpose)  {
	    imagePtr->data = transposeByteArray(imagePtr->data, 
						imagePtr->rows,
						imagePtr->cols);
	    tmp = imagePtr->rows; imagePtr->rows = imagePtr->cols;
	    imagePtr->cols = tmp;
	}
	if(imagePtr->flipx) 
	    imagePtr->data = flipxByteArray(imagePtr->data, 
					    imagePtr->rows,
					    imagePtr->cols);
	if(imagePtr->flipy) 
	    imagePtr->data = flipyByteArray(imagePtr->data, 
					    imagePtr->rows,
					    imagePtr->cols);
    }
    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * ShmImageConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a image widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for imagePtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ShmImageConfigure(Tcl_Interp *interp,	/* Used for error reporting. */
	       ShmImage *imagePtr, 	/* Information about widget. */
	       int argc, 		/* Number of valid entries in argv. */
	       char **argv, 		/* Arguments. */
	       int flags)		/* Flags to pass to
					 * Tk_ConfigureWidget. */
{
    int needNewData = (imagePtr->data==NULL)?1:0;

    if (Tk_ConfigureWidget(interp, imagePtr->tkwin, configSpecs,
	    argc, argv, (char *) imagePtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Set the background for the window and create a graphics context
     * for use during redisplay.
     */

    Tk_SetWindowBackground(imagePtr->tkwin,
	    Tk_3DBorderColor(imagePtr->bgBorder)->pixel);
    if ((imagePtr->gc == None)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	imagePtr->gc = Tk_GetGC(imagePtr->tkwin,
		GCFunction|GCGraphicsExposures, &gcValues);
    }

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */

    Tk_GeometryRequest(imagePtr->tkwin, imagePtr->width, imagePtr->height);
    Tk_SetInternalBorder(imagePtr->tkwin, imagePtr->borderWidth);

    if(imagePtr->colors == NULL) {
	imagePtr->colors = initColors(&(imagePtr->ncolors), 
				      imagePtr->interp, imagePtr->tkwin);
    }

/*  return (imagePtr->colors != NULL && imagePtr->data != NULL)?TCL_OK:TCL_ERR;
 */

    if(needNewData)  {
	ShmImageReload(interp, imagePtr);
    }

    if (!imagePtr->updatePending) {
	Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	imagePtr->updatePending = 1;
    }

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ShmImageEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on images.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
ShmImageEventProc(ClientData clientData,	/* Information about window. */
	       XEvent *eventPtr)	/* Information about event. */
{
    ShmImage *imagePtr = (ShmImage *) clientData;

    if (eventPtr->type == Expose) {
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if (eventPtr->type == ConfigureNotify) {
	if (!imagePtr->updatePending) {
	    Tk_DoWhenIdle(ShmImageDisplay, (ClientData) imagePtr);
	    imagePtr->updatePending = 1;
	}
    } else if (eventPtr->type == DestroyNotify) {
	Tcl_DeleteCommand(imagePtr->interp, Tk_PathName(imagePtr->tkwin));
	imagePtr->tkwin = NULL;
	if (imagePtr->updatePending) {
	    Tk_CancelIdleCall(ShmImageDisplay, (ClientData) imagePtr);
	}
	Tk_EventuallyFree((ClientData) imagePtr, (Tcl_FreeProc*)ShmImageDestroy);
    }
}

static void fillCell(XImage *img, int x, int y, int w, int h, unsigned long px)
{   /* Fill a rectangle in an ximg with a pixel value */
    int i, j;
    unsigned long *dptr, *dp2;
    
    dptr = (unsigned long *)(img->data + (x+img->xoffset)*(img->bits_per_pixel/8)
			               + (y*img->bytes_per_line));
/*     fprintf(stderr, "x=%d y=%d bpl=%d\n", x, y, img->bytes_per_line);
       fprintf(stderr, "data=0x%lx, start=0x%lx\n", img->data, dptr);
 */
    for(i = 0; i<h; ++i) {
	dp2 = dptr;
	for(j = 0; j < w; ++j) {
	    *dp2++ = px;
	}
	dptr = (unsigned long *)(img->bytes_per_line + (char *)dptr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ShmImageDisplay --
 *
 *	This procedure redraws the contents of a image window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
ShmImageDisplay(ClientData clientData)	/* Information about window. */
{
    ShmImage *imagePtr = (ShmImage *) clientData;
    Tk_Window tkwin = imagePtr->tkwin;
    int w,h,v;

    imagePtr->updatePending = 0;
    if (!Tk_IsMapped(tkwin)) {
	return;
    }

    /* Create a pixmap for double-buffering, if necessary. */
    w = Tk_Width(tkwin); h = Tk_Height(tkwin);
    if(imagePtr->image == 0 || imagePtr->pmw != w || imagePtr->pmh != h) {
	if(imagePtr->image != 0) {
#ifdef SHM
	    XShmDetach(Tk_Display(tkwin), &imagePtr->shminfo);
#endif /* SHM */
	    XDestroyImage(imagePtr->image);
	    imagePtr->image = 0;
#ifdef SHM
	    shmdt(imagePtr->shminfo.shmaddr);
	    shmctl(imagePtr->shminfo.shmid, IPC_RMID, 0);
#else /* !SHM */
	    ckfree(imagePtr->actdata);
#endif /* SHM */
	}
	int format = ZPixmap;
#ifdef SHM
	imagePtr->image = XShmCreateImage(Tk_Display(tkwin), 
					  Tk_Visual(Tk_Window(tkwin)), 
				       DefaultDepthOfScreen(Tk_Screen(tkwin)), 
					  format, NULL, 
					  &(imagePtr->shminfo), w, h);
#else /* !SHM */
	imagePtr->image = XCreateImage(Tk_Display(tkwin), 
				       Tk_Visual(Tk_Window(tkwin)),
				       DefaultDepthOfScreen(Tk_Screen(tkwin)),
				       format, 0, NULL, w, h, 32, 0);
#endif /* SHM */
	if (imagePtr->image == 0) {
	    fprintf(stderr, "Error: XShmCreateImg returned %lx\n", 
		    imagePtr->image);
	    return;
	}
	fprintf(stderr, "XImg: w=%d h=%d xof=%d fmt=%d by_o=%d bmp_u=%d bmp_bo=%d bmp_p=%d dpth=%d by_pl=%d bpp=%d red=0x%lx grn=0x%lx blu=0x%lx\n", 
		imagePtr->image->width, imagePtr->image->height, 
		imagePtr->image->xoffset, imagePtr->image->format, 
		imagePtr->image->byte_order, imagePtr->image->bitmap_unit,
		imagePtr->image->bitmap_bit_order, imagePtr->image->bitmap_pad,
		imagePtr->image->depth, imagePtr->image->bytes_per_line, 
		imagePtr->image->bits_per_pixel, imagePtr->image->red_mask, 
		imagePtr->image->green_mask, imagePtr->image->blue_mask);
  

#ifdef SHM
	/* Create the shared memory segment */
	imagePtr->shminfo.shmid = shmget(IPC_PRIVATE, 
					 imagePtr->image->bytes_per_line 
					 * imagePtr->image->height, 
					 IPC_CREAT|0777);
	fprintf(stderr, "shmid=%d\n", imagePtr->shminfo.shmid);
	if(imagePtr->shminfo.shmid == -1) {
	    fprintf(stderr, "Error: shmget returned %d\n", 
		    imagePtr->shminfo.shmid);
	    XDestroyImage(imagePtr->image);
	    imagePtr->image = 0;
	    return;
	}
	imagePtr->shminfo.shmaddr = 
	    imagePtr->image->data = (char*)shmat(imagePtr->shminfo.shmid,0,0);
	if(imagePtr->shminfo.shmaddr == (char*)-1) {
	    fprintf(stderr, "Error: shmat returned %d\n", 
		    imagePtr->shminfo.shmaddr);
	    XDestroyImage(imagePtr->image);
	    imagePtr->image = 0;
	    shmctl(imagePtr->shminfo.shmid, IPC_RMID, 0);
	    return;
	}
	imagePtr->shminfo.readOnly = False;	/* can do XShmGetImage FWIW */
	Status s = XShmAttach(Tk_Display(tkwin), &imagePtr->shminfo);
	if(s == 0) {
	    fprintf(stderr, "XShmAttach returned %d\n", s);
	    XDestroyImage(imagePtr->image);
	    imagePtr->image = 0;
	    shmdt(imagePtr->shminfo.shmaddr);
	    shmctl(imagePtr->shminfo.shmid, IPC_RMID, 0);
	    return;
	}		
#else /* !SHM */
	imagePtr->actdata = (char*)ckalloc(imagePtr->image->bytes_per_line 
					       * imagePtr->image->height + 4);
	imagePtr->image->data = (char*)(((long int)(imagePtr->actdata))&-4);
#endif /* SHM */
	fprintf(stderr, "data=0x%lx\n", imagePtr->image->data);
	imagePtr->pmw = w;
	imagePtr->pmh = h;
	imagePtr->pmCbase = 0;
	imagePtr->pmClen  = 0;
	/* Redraw the widget's background and border only once per pm. */
	/* Tk_Fill3DRectangle(tkwin, imagePtr->pm, 
	   imagePtr->bgBorder,
	   0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
	   imagePtr->borderWidth, imagePtr->relief); */
    }

    /* Display the image. */
    /* first, set up work area ptr (->x, ->xsize etc) */
    ShmImageGetWorkArea(imagePtr);
    /* problem now is to fit */
    int x1, x2, y1, y2;
    int i,j;
    int arxmin, arxmax, arymin, arymax;
    float arw, arh;

    /* one Fill command for each data array element */
    arxmin = (int)imagePtr->cbase;
    if( (arw = imagePtr->clen) == 0) {
	arw = imagePtr->cols - imagePtr->cbase;
    }
    arxmax = (int)(0.99+imagePtr->cbase+arw);
/* 	fprintf(stderr, "redrawing cols %d..%d (data %dx%d)\n", 
	arxmin, arxmax,imagePtr->rows, imagePtr->cols);	*/
    arymin = (int)imagePtr->rbase;
    if( (arh = imagePtr->rlen) == 0) {
	arh = imagePtr->rows - imagePtr->rbase;
    }
    arymax = (int)(0.99+imagePtr->rbase+arh);

    for(i = arxmin; i<arxmax; ++i)  {
	for(j = arymin; j<arymax; ++j)  {
	    x1 = imagePtr->x 
		+ imagePtr->xsize*(((float)i-imagePtr->cbase)/arw);
	    if(x1 < imagePtr->x)	x1 = imagePtr->x;
	    x2 = imagePtr->x 
		+ imagePtr->xsize*(((float)i+1.0-imagePtr->cbase)/arw);
	    if(x2 > imagePtr->x + imagePtr->xsize)
		x2 = imagePtr->x + imagePtr->xsize;
	    /* y1 is *lower* edge, meaning *larger* val of y */
	    y1 = imagePtr->y + imagePtr->ysize 
		- imagePtr->ysize*(((float)j-imagePtr->rbase)/arh);
	    if(y1 > imagePtr->y + imagePtr->ysize)
		y1 = imagePtr->y + imagePtr->ysize;
	    y2 = imagePtr->y  + imagePtr->ysize  
		- imagePtr->ysize*(((float)j+1.0-imagePtr->rbase)/arh);
	    if(y2 < imagePtr->y)	y2 = imagePtr->y;
	    /* y2 is supposed to be the smaller of the two */
	    if(y2 > y1)	y2 = y1;
	    /* data stored with cols most rapid dimension */
	    v = imagePtr->data[i + imagePtr->cols*j];
	    v = (v*imagePtr->ncolors)/256;
	    fillCell(imagePtr->image, x1, y2, x2-x1, y1-y2, 
                     imagePtr->colors[v]->pixel);
	}
	imagePtr->pmCbase = imagePtr->cbase;
	imagePtr->pmClen  = arw /* imagePtr->clen */ ;
    }
#ifdef SHM
    Status s = XShmPutImage(Tk_Display(tkwin), Tk_WindowId(tkwin), 
	      imagePtr->gc, imagePtr->image, 0, 0, 
	      0, 0, Tk_Width(tkwin), Tk_Height(tkwin), False);
#else /* !SHM */
    Status s = XPutImage(Tk_Display(tkwin), Tk_WindowId(tkwin), 
	      imagePtr->gc, imagePtr->image, 0, 0, 
	      0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
#endif /* SHM */
    if(s == 0) {
	fprintf(stderr, "XShmPutImage returned %d\n", s);
    }		
}

/*
 *----------------------------------------------------------------------
 *
 * ShmImageDestroy --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a image at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the image is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
ShmImageDestroy(ClientData clientData)	/* Info about image widget. */
{
    ShmImage *imagePtr = (ShmImage *) clientData;
    int i;

    Tk_FreeOptions(configSpecs, (char *) imagePtr, imagePtr->display, 0);
    if (imagePtr->gc != None) {
	Tk_FreeGC(imagePtr->display, imagePtr->gc);
    }
    if(imagePtr->colors) {
	for(i = 0; i<imagePtr->ncolors; ++i)  
	    Tk_FreeColor(imagePtr->colors[i]);
	ckfree((char *)imagePtr->colors);
	imagePtr->colors = NULL;
	imagePtr->ncolors = 0;
    }
    if(imagePtr->data)  {
	ckfree((char *)imagePtr->data);
	imagePtr->data = NULL;
    }
    ckfree((char *) imagePtr);
}
