// -------------------------------------------------------------------
// tkWaveform.cc
//
// tk widget to display any old waveform
// dpwe@mit.edu 1994sep14
//

#include <assert.h>
#include <math.h>	/* for atof() */

extern "C" {
/* INT32 is redefined by X11/Xmd.h (on my IRIX405H) which CC doesn't 
   like, so unset it here */
#undef INT32
#include "tcl.h"
#include "tk.h"

#include <vecops.h>	/* for floatVector handling */

}

// Widget data structure

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    XColor	*selColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    // data for the drawing activities
    GC  	gc;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    // fields describing the content
    int		bufLen;		// how many points into the past do we hold?
    int		bufConts;	// how many valid points actually in there?
    float	*buf;		// buffer containing the data
    double	scale;		// scaling for data onto screen
} Waveform;

extern "C" {
// static function prototypes

int Waveform_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int WaveformCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  WaveformConfigure(Tcl_Interp *interp, Waveform *sw, 
			     int argc, char **argv, int flags);
static void WaveformDestroy(ClientData clientData);
static void WaveformDisplay(ClientData clientData);
static void WaveformEventProc(ClientData clientData, XEvent *eventPtr);
static int  WaveformWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);
};

// Argv parsing info

static Tk_ConfigSpec WaveformConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#d9d9d9", Tk_Offset(Waveform, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "#404040", Tk_Offset(Waveform, fgColor), 0 }, 
    {TK_CONFIG_COLOR, "-selColor", "selColor", "SelColor", 
     "#A0A0A0", Tk_Offset(Waveform, selColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(Waveform, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(Waveform, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(Waveform, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(Waveform, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(Waveform, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(Waveform, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(Waveform, doubleBuffer), 0},
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale",
	"1.0", Tk_Offset(Waveform, scale), 0},
    {TK_CONFIG_INT, "-buflen", "buflen", "Buflen",
	"0", Tk_Offset(Waveform, bufLen), 0},
    {TK_CONFIG_INT, "-bufconts", "bufconts", "Bufconts",
	"0", Tk_Offset(Waveform, bufConts), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int Waveform_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "waveform", WaveformCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int WaveformCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new Waveform widget */
    Tk_Window	main = (Tk_Window) clientData; 
    Waveform 	*sw; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Waveform");

    // Allocate and initialize the structure
    sw = (Waveform *)ckalloc(sizeof(Waveform));
    sw->tkwin = tkwin;
    sw->display = Tk_Display(tkwin);
    sw->interp  = interp;
    sw->bgBorder= NULL;
    sw->fgColor = NULL;
    sw->selColor = NULL;
    sw->gc = None;
    sw->doubleBuffer = 1;
    sw->updatePending = 0;

    sw->scale= 1.0;
    sw->bufLen = 0;
    sw->bufConts = 0;
    sw->buf = NULL;

    Tk_CreateEventHandler(sw->tkwin, ExposureMask|StructureNotifyMask,
			  WaveformEventProc, (ClientData)sw);
    Tcl_CreateCommand(interp, Tk_PathName(sw->tkwin), WaveformWidgetCmd, 
		      (ClientData)sw, (void (*)(ClientData))NULL);
    if(WaveformConfigure(interp, sw, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(sw->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(sw->tkwin);
    return TCL_OK;
}

static void WaveformGetWorkArea(Waveform *sw, int *px, int *py, int *pw, int *ph)
{   /* return the area that waveform is actually drawn in */
    Tk_Window tkwin = sw->tkwin;

    *px = sw->padX;
    *pw = Tk_Width(tkwin) - 2*sw->padX;
    *py = sw->padY;
    *ph = Tk_Height(tkwin) - 2*sw->padY;
}

static int findOneNumber(char **p, double *ret)
{   /* passed a modifiable pointer to a string, skip any whitespace 
       and attempt to parse a float that comes next.  If success, 
       return 1 and write value into ret.  If hit end of string without 
       a number, return 0.  If string is not a number, return -1.
       <p> is modified ready for next call. */
    char   *r, *s=*p;
    double val;

    /* skip leading ws */
    s += strspn(s, " \n\r\t");
    if(*s=='\0')  {	/* hit end of string */
	return 0;
    }
    val = strtod(s, &r);
    if(r==s) {	/* no conversion was performed - bad string? */
	return -1;
    }
    *ret = val;
    *p = r;
    return 1;
}

static int WaveformParseData(Tcl_Interp *interp, Waveform *sw, char *data)
{   /* data is a string containing several new numbers to push into 
       the data.  See how many there are and incorporate them. */
    char *s = data;
    int i, n = 0, rc;
    double val;
    float *bptr;

    /* parse whole string once to count number of elements and 
       check syntax */
    while( (rc=findOneNumber(&s, &val)) == 1)  {
	++n;
    }
    if(rc < 0)  {
	Tcl_AppendResult(interp, "cannot parse as number :",
			 s, NULL);
	return TCL_ERROR;
    }
/*    fprintf(stderr, "parseData: %d data passed\n", n);   */
    /* OK, parse the string again, shift new data into our buffer */
    s = data;
    if(n > sw->bufLen)  {
	int discard = n - sw->bufLen;
	fprintf(stderr, "parseData: discarding %d of %d data\n", discard, n);
	for(i=0; i<discard; ++i)
	    findOneNumber(&s, &val);	/* just parse over them */
	n -= discard;
    }
    /* How many of the existing buffer contents will be pushed out 
       by the new data?  As many as are required to respect bufLen */
    int toBeDropped  = MAX(0, n+sw->bufConts - sw->bufLen);
    /* so how many of the existing contents will we keep? */
    int goodOldConts = sw->bufConts - toBeDropped;
    /* what will the contents be after this filling? */
    int newBufConts = goodOldConts + n;
    /* The new items will go at the end of the buffer.  Copy back 
       the existing contents to make room, if necessary */
    if(toBeDropped > 0 && goodOldConts > 0) {
	VCopy(sw->buf+toBeDropped, 1, sw->buf, 1, goodOldConts);
    }
    bptr = sw->buf+goodOldConts;	/* base for new data */
    for(i=0; i<n; ++i)  {
	rc = findOneNumber(&s, &val);
	assert(rc==1);			/* since we already parsed it */
	*bptr++ = val;
    }
    sw->bufConts = newBufConts;
    return TCL_OK;
}

static int WaveformWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    Waveform	*sw = (Waveform *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)sw);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, WaveformConfigSpecs, 
				      (char *)sw, NULL, 0);
	} else if(argc==3) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, WaveformConfigSpecs, 
				      (char *)sw, argv[2], 0);
	} else {
	    result = WaveformConfigure(interp, sw, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, sw->tkwin, WaveformConfigSpecs,
				   (char *) sw, argv[2], 0);
    } else if( (c == 'p') && (strncmp(argv[1], "push", length)==0) )  {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " push <datalist>\"", NULL);
	    return TCL_ERROR;
	}
        // argv3 is a bunch of numbers.  Maybe scan first
	result = WaveformParseData(interp, sw, argv[2]);
    } else if( (c == 'd') && (strncmp(argv[1], "dump", length)==0) )  {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " dump\"", NULL);
	    return TCL_ERROR;
	}
	int i;
	char s[32];
	for(i=0; i<sw->bufConts; ++i)  {
	    sprintf(s, "%f ", sw->buf[i]);
	    Tcl_AppendResult(interp, s, (char *)NULL);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, cget, push or dump", (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !sw->updatePending) {
	Tk_DoWhenIdle(WaveformDisplay, (ClientData) sw);
	sw->updatePending = 1;
    }
    Tk_Release((ClientData)sw);
    return result;
}

static int WaveformConfigure(Tcl_Interp *interp, Waveform *sw, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;
    int origLen = sw->bufLen;

    if(Tk_ConfigureWidget(interp, sw->tkwin, WaveformConfigSpecs, 
			  argc, argv, (char *)sw, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    // check for change of len
    if(sw->bufLen != origLen)  {
	float *newbuf = (float *)ckalloc(sw->bufLen*sizeof(float));
	if(sw->buf)  {	/* i.e. there is some old data */
	    int tokeep;
	    tokeep = MIN(sw->bufConts, sw->bufLen);
	    // copy across data to keep
	    VCopy(newbuf, 1, sw->buf+sw->bufConts-tokeep, 1, tokeep);
	    sw->bufConts = tokeep;
	    ckfree((char *)sw->buf);
	}
	sw->buf = newbuf;
    }
    // set the colors and GC for the redisplay
    Tk_SetWindowBackground(sw->tkwin, Tk_3DBorderColor(sw->bgBorder)->pixel);
    if ((sw->gc == None) && (sw->doubleBuffer)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	sw->gc = Tk_GetGC(sw->tkwin,GCFunction|GCGraphicsExposures, &gcValues);
    }
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(sw->tkwin, sw->width, sw->height);
    Tk_SetInternalBorder(sw->tkwin, sw->borderWidth);
    if (!sw->updatePending) {
	Tk_DoWhenIdle(WaveformDisplay, (ClientData)sw);
	sw->updatePending = 1;
    }
    return rslt;
}

static void WaveformEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the Waveform widget */
    Waveform *sw = (Waveform *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(WaveformDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(WaveformDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(sw->interp, Tk_PathName(sw->tkwin));
	sw->tkwin = NULL;
	if(sw->updatePending)  {
	    Tk_CancelIdleCall(WaveformDisplay, (ClientData)sw);
	}
	Tk_EventuallyFree((ClientData)sw, (Tcl_FreeProc*)WaveformDestroy);
    }
}

static void WaveformDrawWvfm(Waveform *sw, int xmin, int xmax, Drawable d=None)
{   /* actually draw waveform data from our buf */
    Tk_Window 		tkwin = sw->tkwin;
    GC 			gc    = sw->gc;
    Display 		*disp = Tk_Display(tkwin);
    XColor 		*fgColor = sw->fgColor;
    double 		scale   = sw->scale;
    int x, y, w, h, y0, yscale, xd=0, yd=0, xo, yo;
    int i;
    int nxs, xssize = 100;
    XSegment *xsgs, *thisSeg;
    float	*bptr = sw->buf;
    int		conts= sw->bufConts;
    int		len  = sw->bufLen;
    int 	offs = len-conts;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    WaveformGetWorkArea(sw, &x, &y, &w, &h);

    y0 = y+h/2;
    yscale = -h/2;
    
    XSetForeground(disp, gc, fgColor->pixel);
    
    xsgs = (XSegment *)ckalloc(xssize * sizeof(XSegment));
    thisSeg = xsgs;
    nxs = 0;
    for(i = 0; i < conts; ++i)  {
	yo = yd;  xo = xd;
	yd = int(y0+yscale*scale**bptr++);
	xd = x+(((offs+i)*w)/(len-1));	// len-1 because that's the # of segs
	if(i>0)  {
	    thisSeg->x1 = xo;	thisSeg->y1 = yo;
	    thisSeg->x2 = xd;	thisSeg++->y2 = yd;
	    if(++nxs == xssize)  {
		XDrawSegments(disp, d, gc, xsgs, nxs);
		thisSeg = xsgs;
		nxs = 0;
	    }
	}
    }
    if(nxs > 0)  {
	XDrawSegments(disp, d, gc, xsgs, nxs);
    }
    ckfree((char *)xsgs);
}

static void WaveformDisplay(ClientData clientData)
{				/* actually draw the contents of the widget */
    Waveform *sw = (Waveform *)clientData;
    Tk_Window tkwin = sw->tkwin;
    Pixmap pm = None;
    Drawable d;
    Display *disp = sw->display;
    int 	tw, th, inset;
    
    sw->updatePending = 0;
    if(!Tk_IsMapped(tkwin)) {
	return;
    }
    
    tw = Tk_Width(tkwin);
    th = Tk_Height(tkwin);
    
    // Create pixmap for double-buffering if necessary
    if(sw->doubleBuffer)  {
	pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
	d = pm;
    } else {		// draw directly into window
	d = Tk_WindowId(tkwin);
    }
    // redraw background
    Tk_Fill3DRectangle(tkwin, d, sw->bgBorder, 0, 0, tw, th, sw->borderWidth, 
		       sw->relief);
    // Actually do the drawing
    inset = 2;
    
    //    drawX(tkwin, sw->gc, d, sw->borderWidth + inset);
    WaveformDrawWvfm(sw, 0, 0, d);
    
    // copy to screen if double-buffering
    if(sw->doubleBuffer)  {
	XCopyArea(disp, pm, Tk_WindowId(tkwin), sw->gc, 0, 0, 
		  tw, th, 0, 0);
	XFreePixmap(disp, pm);
    }
}

static void WaveformDestroy(ClientData clientData)
{				/* destroy one of our objects */
    Waveform *sw = (Waveform *)clientData;
    
    Tk_FreeOptions(WaveformConfigSpecs, (char *)sw, sw->display, 0);
    if(sw->gc != None)  {
	Tk_FreeGC(sw->display, sw->gc);
    }
    // don't destroy sound as we didn't new it
    ckfree((char *)sw);
}


