#!./sowish -file
#!/usr/local/bin/sowish -file
# tsoundg
# first stab at a soundfile viewer under tk/tck
# Uses extensions: sound (tclSound.cc), sndwvfm (tkSndWvfm.cc)
#		   timeline (tkTimeline.cc), image (tkImg.c)

if { $argc < 1 } {
    puts "usage: tsoundg <soundfile> \[<graphDATfile>\]"
    exit -1
}

# Make sure libtksound is loaded if this is plain wish
if {[info commands sound] == ""} {
    set libfile $env(HOME)/lib/$env(SPEECH_ARCH)/libtksound[info sharedlibext]
    if {[file readable $libfile] && [info commands load] != ""} {
	load $libfile
    } else {
	puts stderr "$argv0: cannot load library $libfile (Tcl version is $tcl_version)"
	exit -1
    }
}

source dpweutils.tcl

# allow terminal inputs as well
set tcl_interactive 1

# make sure the window is resizable
wm minsize . 1 1

# load the sound
set filename [lindex $argv 0]
set sound so
sound $sound
$sound open $filename
# $sound play

# find the filename for the graphics (spectrogram) file
if { $argc < 2 } {
  # if not specified, default to <root>.dat
  regsub {\.[^.]*$} $filename ".dat" hexname
  #regsub {\.[^.]*$} $filename ".hex" hexname
} else {
  set hexname [lindex $argv 1]
}

proc Open {filename {mode "r"}} {
    # Like open, but returns failures as null tokens rather than 
    # crashing out
    if {[catch {set file [open $filename $mode]} err]} {
	return ""
    } else {
	return $file
    }
}

proc ReadLabelFile {filename {warn 1}} {
    # Reads a file in NIST/ESPS labelled format into a 
    # list whose first element is the header lines (separated by \n's)
    # and whose second element is a list of pairs like
    # {time {rest of token line}}
    if {[set file [Open $filename "r"]]==""} {
	Error "Unable to read '$filename' as label file"
	return {}
    }
    # Read each line
    set hdr {}
    set inhdr 1
    set rslt {}
#    set delpattern [format {*%c*} 127]
    set delpattern [format {*[%c%c-%c%c-%c]*} 127 1 8 10 31]
    while {![eof $file]} {
    	set line [string trim [gets $file]]
	# Escape any stray braces that will mess up the list handling
	regsub -all "\{" $line "\\\{" line
	regsub -all "\}" $line "\\\}" line
	regsub -all \\\[ $line "\\\[" line
	regsub -all \\\] $line "\\\]" line
	regsub -all "\"" $line "\\\"" line
	if {[regsub -all ";" $line "\\;" line] && !$inhdr && $warn} {
	    puts stderr "WARNING: semicolon in $filename at $line"
	}
	# Check for weird characters	
	if {[string match $delpattern $line]} {
	    puts stderr "$filename: odd chrs in '$line'"
	}
    	set c [string index $line 0]
        if {$inhdr} {
	    # gather the header lines more-or-less verbatim
	    append hdr "$line\n"
	    if {$c == "#"} {
		# End of header is marked by a single hash on its own
		set inhdr 0
	    }
	} else {
	    # Past the header - treat lines as tokens
	    if {[llength $line]>0 && $c != "#"} {
		lappend rslt [list [lindex $line 0] [lrange $line 1 e]]
	    }
	}
    }
    close $file
    return [list $hdr $rslt]
}

proc LabelsToDurs {labs {filter 1} {reportEmpty 1}} {
    # Convert the label structure read from a file into things with durations
    # i.e. {etime colr tok} {etime colr tok} becomes
    # {btime dur tok} {btime dur tok}
    # Specifying that $filter is zero will prevent the stripping of h#s 
    # from the output
    set durlabs {}
    # Assume first label comes from start of file
    set lasttime 0
    foreach lab $labs {
	# Break out the label fields
	set time [lindex $lab 0]
	# Convert $time to Tcl-valid float repn?
	set time [expr $time + 0.0]
	set colr [lindex [lindex $lab 1] 0]
	set tok  [lrange [lindex $lab 1] 1 e]
	# Skip H#s or whatever
	set TOK [string toupper $tok]
#	if {$colr != 121} {
#	    puts stderr "** color=$colr, not 121"
#	}
	# Don't ignore empty labels for now
        if {!$filter || \
		($TOK != {} && $TOK != "SIL" && $TOK != "H#" \
		 && $TOK != "UNL" && $TOK != "SP")} {
	    lappend durlabs [list $lasttime [expr $time-$lasttime] $tok]
	} elseif {$TOK == {} && $reportEmpty} {
	    puts "**Empty TOK in $labs"
	}
	set lasttime $time
    }
    return $durlabs
}

# find the filename for the labels (phoneme) file
regsub {\.[^.]*$} $filename ".phn" labname
# load the labels
if {[file readable $labname]} {
    set labs [LabelsToDurs [lindex [ReadLabelFile $labname] 1]]
} else {
    set labs 0
}

proc DrawLabs {labs cnv timescale {timebase 0.0} {maxx 9999}} {
    # Draw the labels in $labs on the canvas $cnv
    # using the timescale of $timescale pixel/sec
    # with pixel zero meaning time $timebase
    global xpad
    set tag "labs"
    $cnv delete $tag
    set xfix 2
    set ybase 0
    set lastx 9999
    set lasty 0
    foreach lab [reverse $labs] {
	lassign $lab time dummy tok
	set x [expr $xpad + int(($time - $timebase)*$timescale)]
	set y $ybase
	if {$x > $xpad && $x < ($maxx+$xpad)} {
	    lassign [CanvTextDims $cnv $tok] wdth hght
	    if {[expr $x+$wdth] >= $lastx} {
		# Overlap with preceding label (right to left) - move down
		# (if it would fit)
		if {[expr $lasty+2*$hght] <= [WinHeight $cnv]} {
		    set y [expr $lasty + $hght]
		}
	    }
	    $cnv create text [expr $x+$xfix] $y -anchor nw -text $tok -tag $tag
	    # Add lines
	    set bottom [expr $y+$hght]
	    set right [expr $x+$wdth]
	    $cnv create line $x $ybase $x $bottom $right $bottom -tag $tag
	    # Remember position of this label
	    set lastx $x
	    set lasty $y
	}
    }
}

######### Widget callback functions ###########

proc timelineAndScrollCommand {timeline scrollbar image sound \
				totConts viewSize top bottom} {
    # normally the soundview scrollback command plugs straight into 
    # the scrollbar - but here we steal off a bit to feed the time line 
    # as well.
    # Also, we can take this opportunity to scale the `increment' used 
    # for the end-arrows in the scrollbar.
#    puts "$totConts $viewSize $top $bottom"
#   e.g 52941 10000 8100 18100
    global scrollScale
    if { $top > $totConts } {
	set top $totConts
    }
#    if { $top < -$viewSize } {
#	set top [expr "-$viewSize"]
#    }
#    if { $top < 0 } {
#	set top 0
#    }
    set bottom [expr "$top+$viewSize"]
    # that's what I actually want
    set srate [$sound cget -samplerate]
    set tmin  [expr "$top / $srate"]
    set tmax  [expr "$bottom/$srate"]
    # set the timeline to display the new range
    $timeline config -minVal $tmin -maxVal $tmax
    # now move the scrollbar thumb to the right place
    # scroll bar operates in units 1/16th of the total short view width
    set scrollScale [expr "$viewSize / 200"]
    $scrollbar set [expr "$totConts/$scrollScale"] \
		   [expr "$viewSize/$scrollScale"] \
		   [expr "$top/$scrollScale"] [expr "$bottom/$scrollScale"]
    # also update the image to display the right range of columns
    set icols [expr [$image cget -cols]+0.0]
    set ibase [expr "$icols * $top / $totConts"]
    set ilen  [expr "$icols * $viewSize / $totConts"]
    $image config -cbase $ibase -clen $ilen

    global labs xpad
    set cnv ".labs"
#    if {[info command $cnv] != ""} {
	#set w [$cnv cget -width]
    	set w [expr [WinWidth $cnv] - 2*$xpad]
	DrawLabs $labs $cnv [expr $w/($tmax-$tmin)] $tmin $w
#    }
}

proc SndNewView {snd base {len 0}} {
    # Set a new view on a sndwvfm $snd
    if {$base < 0} {set base 0}
    if {$len <=0} {set len [$snd cget -viewLen]}
    $snd config -viewBase $base -viewLen $len
}

proc scrollbarCommand {soundview baseval} {
    # quick filter to take output from scrollbar callback and scale to samples
    global scrollScale
    SndNewView $soundview [expr "$baseval * $scrollScale"]
}
proc nosoundScrollbarCommand {soundview baseval} {
    # scrollback command that just scrolls the spectrogram, not the shortview
    # scanning the soundfile is just too slow
    global scrollScale
    set bv [expr "$baseval * $scrollScale"]
    set sound [$soundview cget -sound]
    set tc [$sound cget -nframes]
    set vs [$soundview cget -viewLen]
    timelineAndScrollCommand .swtl .sb .im $sound $tc $vs $bv \
	                       [expr "$bv + $vs"]
}

# xpad is supposed to be size of scrollbar arrows
set xpad 16
# 16 if scrollbar relief is sunken
set soDuration [expr "[$sound cget -nframes]/[$sound cget -samplerate]"]

farray fa
fa load $hexname

##########  Initialize all the widgets #############
# Canvas for adding labels - create first so exists after 1st scrollback callb
canvas    .labs   -height 48

image     .im     -farray fa -padX $xpad -padY 5 -relief flat
.im config  -rlen [.im cget -rows]
label     .status -text "File: $filename"

sndwvfm   .ov     -sound $sound -padX $xpad -padY 5 -height 100 \
		  -viewLen [$sound cget -nframes]
set sndwvfmStatusArray(.ov) "noscroll"
timeline  .ovtl   -padX $xpad -padY 0 -height 20 -maxVal $soDuration \
		  -tixTopLeft 0 -relief flat
timeline  .swtl   -padX $xpad -padY 0 -height 20 -maxVal $soDuration \
		  -tixTopLeft 1 -relief flat
sndwvfm   .sw     -sound $sound -padX $xpad -padY 5 -viewLen 10000
set sndwvfmStatusArray(.sw) "scrollOK"

# Swap the commenting of the next two lines for faster scrolling 
#   which does not update the shortview until mouseup.
scrollbar .sb     -orient horiz -relief flat -command "scrollbarCommand .sw"
#scrollbar .sb     -orient horiz -relief flat -command "nosoundScrollbarCommand .sw"
# ".sb set"
bind .sb <ButtonRelease-1> 	{scrollbarCommand .sw [lindex [.sb get] 2]}

# drag handles - fillers to allow vertical resizing
frame .drag1 -height 5 -cursor bottom_side
frame .drag2 -height 5 -cursor bottom_side
frame .drag3 -height 5 -cursor bottom_side

# global list of views associated with this sound
set $sound.views ".sw .ov"

frame  .buttons
button .buttons.zoomin  -text " < "  -command "zoom .sw .7"
button .buttons.zoomout -text " > "  -command "zoom .sw 1.4"
button .buttons.zoomsel -text "zoom" -command "zoomsel .sw"
button .buttons.quit    -text "quit" -command "exit"
button .buttons.play    -text "play" -command "svplay .sw"

######### Pack all the widgets ###########
pack .status          -side top -fill x -expand no
pack .buttons         -side bottom -fill x -expand no -padx 6 -pady 6
pack .buttons.zoomin  -side left
pack .buttons.zoomout -side right
pack .buttons.play .buttons.zoomsel .buttons.quit -side left -expand yes

pack .ovtl  -side top    -fill x
pack .ov    -side top    -fill x
pack .drag1 -side top    -fill x

pack .sb    -side top    -fill x
pack .sw    -side top	 -fill x
pack .drag2 -side top    -fill x

# pack .im    -side top    -fill both
# pack .drag3 -side top    -fill x
# pack .swtl  -side top    -fill x
### doesn't allow image panel to grow/shrink when whole window resized

pack .swtl  -side bottom -fill x
pack .drag3 -side bottom -fill x
pack .im    -side bottom -fill both -expand yes

pack .labs  -side bottom -fill x

tkwait visibility .labs
# Now can invoke the first scrollCommand...
.sw config -scrollCommand {timelineAndScrollCommand .swtl .sb .im [.sw cget -sound]}

########## dragging bit

proc startDragProc {widg target x y} {
    # widg is the drag widget, target is the widget whose height is 
    # to track the up/down motion over this widget
    $widg config -cursor sb_v_double_arrow
    set my [expr "$y - [winfo rooty .]"]
    global dragStart$widg startHeight$target
#    set dragStart$widg [lindex [local2global $widg $x $y] 1]
    set dragStart$widg $my
    set startHeight$target [$target cget -height]
#    puts "startDrag: y [set dragStart$widg] h [set startHeight$target]"
    xorrect . 0 $my 2000 3
}

proc runDragProc {widg target x y} {
    global dragStart$widg startHeight$target
    set my [expr "$y - [winfo rooty .]"]
#    puts "runDrag: y $my"
#    $target config -height [expr "[set startHeight$target] + $my - [set dragStart$widg]"]
    xorrect . 0 $my 2000 3
}

proc endDragProc {widg target x y} {
    global dragStart$widg startHeight$target
#    set gy [lindex [local2global $widg $x $y] 1]
    set my [expr "$y - [winfo rooty .]"]
    $widg config -cursor bottom_side
#    puts "runDrag: y $my"
    xorrect .
    $target config -height [expr "[set startHeight$target] + $my - [set dragStart$widg]"]
}

bind .drag1 <1> 			{startDragProc %W .ov %X %Y}
bind .drag1 <B1-Motion> 		{runDragProc   %W .ov %X %Y}
bind .drag1 <ButtonRelease-1> 		{endDragProc   %W .ov %X %Y}

bind .drag2 <1> 			{startDragProc %W .sw %X %Y}
bind .drag2 <B1-Motion> 		{runDragProc   %W .sw %X %Y}
bind .drag2 <ButtonRelease-1> 		{endDragProc   %W .sw %X %Y}

bind .drag3 <1> 			{startDragProc %W .im %X %Y}
bind .drag3 <B1-Motion> 		{runDragProc   %W .im %X %Y}
bind .drag3 <ButtonRelease-1> 		{endDragProc   %W .im %X %Y}


######## Selection and autoscrolling functions ############

proc scroll {sndview percent} {
    # shift the current view by some percentage of the current view - 
    #   forward if $percent > 0, else backward
    set viewBase [$sndview cget -viewBase]
    set viewLen  [$sndview cget -viewLen ]
    set newBase [expr "$viewBase + ($viewLen * $percent)/100"]
    SndNewView $sndview $newBase
}

proc startSel {sndview xpix}  {
    # mouse down in view
    global anchor$sndview
    set frame [$sndview frameOfPixel $xpix]
    # set new selection in all concerned views
    set viewsname [$sndview cget -sound].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selLen 0 -selBase $frame
    }
    set anchor$sndview $frame
}

# semaphore to pre-empt auto-repeats
set sndwvfm_auto_pending 0
set sndwvfm_cancel_auto  0

proc endSel {sndview} {
    # called when the scroll ends: if an auto pending, be sure to cancel it
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    if { $sndwvfm_auto_pending }  {
	set sndwvfm_cancel_auto 1
    }
}

proc scrollGrowSel {sndview scrollamt}  {
    # scroll the view and act as if the button was moved, since it 
    # has moved, relative to the scrolled view.
    # This command is executed by a timer a little while after the 
    # selection is grown into the 'end zones'
    global sndwvfm_auto_pending sndwvfm_cancel_auto currentX$sndview
    set sndwvfm_auto_pending 0
    if { $sndwvfm_cancel_auto } {
	# this autoscroll event was cancelled before it ever got to happen
	set sndwvfm_cancel_auto 0
    } else {
	scroll $sndview $scrollamt
	# re-execute the scroll .. 
	growSel $sndview [set currentX$sndview]
    }
}

proc checkForAuto {sndview xpix}  {
    # if we're near the edge of the view, schedule an auto-scroll
    # ('near' is within $border pixels)
    global sndwvfm_auto_pending sndwvfm_cancel_auto sndwvfmStatusArray
    if {$sndwvfmStatusArray($sndview) == "noscroll"} {return}
    set border 50
    set lowerlim $border
    set upperlim [expr "[winfo width $sndview] - $border"]
    if { $xpix < $lowerlim || $xpix > $upperlim } {
	# calculate the scroll amount so that it depends how far the 
	# mouse is into the autoscroll region
	if { $xpix < $lowerlim } {
	    set xoffset [expr "$xpix - $lowerlim"]
	} else {
	    set xoffset [expr "$xpix - $upperlim"]
	}
	set scrollamt [expr "(5 * $xoffset)/$border"]
	if { $sndwvfm_auto_pending == 0 } {
	    after 200 "scrollGrowSel $sndview $scrollamt"
	    set sndwvfm_auto_pending 1
	}
	# just in case we have cancelled a pending auto, re-instance it
	set sndwvfm_cancel_auto 0
    } else {
	# we have moved out of the autoscroll zone; cancel any pending scroll
	if { $sndwvfm_auto_pending } {
	    set sndwvfm_cancel_auto 1
	}
    }
}

proc growSel {sndview xpix}  {
    # mouse drag in view - just extend the selection
    global anchor$sndview currentX$sndview
    # save the current x position for the redraw to figure
    set currentX$sndview $xpix
    # better set up for an autoscroll if we need one
    checkForAuto $sndview $xpix
    # OK, now just grow the selection
    set origPos [set anchor$sndview]
    set newPos  [$sndview frameOfPixel $xpix]
    if { $newPos < $origPos } {
	set tmp $newPos
	set newPos $origPos
	set origPos $tmp
    }
    set newLen [expr "$newPos - $origPos"]
    # set new selection in all involved windows
    set viewsname [$sndview cget -sound].views
    global $viewsname
    foreach view [set $viewsname] {
       $view config -selBase $origPos -selLen $newLen
    }
}

proc startSelExtend {sndview xpix}  {
    # mouse down in view, but keep more distant of current anchor points
    global anchor$sndview
    set selBase [$sndview cget -selBase]
    set selLen  [$sndview cget -selLen ]
    set selEnd  [expr "$selBase + $selLen"]
    set newPos  [$sndview frameOfPixel $xpix]
    if {$selLen == 0}  {
	# no selection to extend - act like button 1
	startSel $sndview $xpix
    } else {
	# set the anchor to the more distant of the selection points
        if { abs($newPos - $selBase) > abs($newPos - $selEnd) } {
	    set anchor$sndview $selBase
	} else {
	    set anchor$sndview $selEnd
	}
	# and now act as if cursor just moved
	growSel $sndview $xpix
    }
}

####### Bind all the selection and scroll functions #######

bind .sw <1> 			{startSel .sw %x}
bind .sw <B1-Motion> 		{growSel .sw %x}
bind .sw <B1-ButtonRelease> 	{endSel .sw}
bind .sw <3> 			{startSelExtend .sw %x}
bind .sw <B3-Motion> 		{growSel .sw %x}
bind .sw <B3-ButtonRelease> 	{endSel .sw}
# dspB binding - middle button zooms to mark
bind .sw <2> 			{zoomsel .sw}

bind .ov <1> 			{startSel .ov %x}
bind .ov <B1-Motion> 		{growSel .ov %x}
bind .ov <B1-ButtonRelease> 	{endSel .ov}
bind .ov <3> 			{startSelExtend .ov %x}
bind .ov <B3-Motion> 		{growSel .ov %x}
bind .ov <B3-ButtonRelease> 	{endSel .ov}
bind .ov <2> 			{zoomsel .sw}

bind .im <2>			{imgclrStart .im %x %y}
bind .im <B2-Motion>		{imgclrDrag .im %x %y}

############### ImageColor mouse commmands 

set _imgclrScale 1.0
set _imgclrOffset 0.0
set _imgclrX 0
set _imgclrY 0
set _imgclrH 0
set _imgclrW 0

proc imgclrStart {img x y} {
    # Begin adjusting contrast and offset for image
    global _imgclrX _imgclrY _imgclrW _imgclrH _imgclrScale _imgclrOffset
    set _imgclrX $x
    set _imgclrY $y
    set _imgclrScale [$img cget -scale]
    set _imgclrOffset [$img cget -offset]
    set _imgclrW [WinWidth .im]
    set _imgclrH [WinHeight .im]
}

proc imgclrDrag {img x y} {
    global _imgclrX _imgclrY _imgclrW _imgclrH _imgclrScale _imgclrOffset
    set scale  [expr $_imgclrScale*exp(log(2.0)*($x-$_imgclrX)/($_imgclrW/2))]
    set offset [expr $_imgclrOffset+($_imgclrY-$y)/($_imgclrH/2.0)]
    $img config -scale $scale -offset $offset
    $img redraw
}    

##############  Zoom, other button functions ###################

proc sndwvfmSelLims {sndview {noneisall 0}} {
    # return the selection base and len.
    # If $noneisall is present and nonzero, return the entire sound limits 
    # if the selection is zero len
    set selBase  [$sndview cget -selBase]
    set selLen   [$sndview cget -selLen ]
    set viewBase [$sndview cget -viewBase]
    set viewLen  [$sndview cget -viewLen ]
    set sound    [$sndview cget -sound]
    set maxLen   [$sound cget -nframes]
    if { $noneisall && ($selLen == 0) } {
	set selBase 0
	set selLen $maxLen
    }
    return [list $selBase $selLen $viewBase $viewLen $maxLen]
}

proc zoom {sndview factor} {
    # adjust the display length of a sndview by the specified factor
    # if factor > 1.0, zooming out
    # attempt to keep view centered at same point
    set sellims [sndwvfmSelLims $sndview 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set newLen   [expr "int($factor * $viewLen)"]
    set newBase  [expr "$viewBase + ($viewLen - $newLen)/2"]
    # newBase may go negative, but $sndview config will clip it nicely
    SndNewView $sndview $newBase $newLen
}

proc zoomsel {sndview} {
    # zoom sound view to current selection - or whole sound if none
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set maxLen   [lindex $sellims 4]
    # if we're already zoomed to sel, zoom out to full (toggle)
    if { ($viewBase == $selBase) && ($viewLen == $selLen) } {
	set selBase 0
	set selLen $maxLen
    }
    SndNewView $sndview $selBase $selLen
}

proc svplay {sndview} {
    # play the selected region of the soundview
    set sound   [$sndview cget -sound]
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    $sound play $selBase $selLen
}

